import numpy as np
import matplotlib.pyplot as plt
from cpymad.madx import Madx
import xtrack as xt

mad_b2 = Madx()
mad_b2.call("../../test_data/hllhc15_noerrors_nobb/sequence.madx")
mad_b2.use(sequence="lhcb2")
twb2mad = mad_b2.twiss()
summb2mad = mad_b2.table.summ

mad_b4 = Madx()
mad_b4.call("../../test_data/hllhc15_noerrors_nobb/sequence_b4.madx")
mad_b4.use(sequence="lhcb2")
twb4mad = mad_b4.twiss()
summb4mad = mad_b4.table.summ

line_b4 = xt.Line.from_madx_sequence(
    mad_b4.sequence["lhcb2"],
    # deferred_expressions=True
)
line_b4.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7000e9)


line_b4.build_tracker()
twb4xt = line_b4.twiss()


# Survey with offset:
starting = {
    "theta0": -np.pi / 9,
    "psi0": np.pi / 7,
    "phi0": np.pi / 11,
    "X0": -300,
    "Y0": 150,
    "Z0": -100,
}
#starting = {}

survb4mad = mad_b4.survey(**starting).dframe()
survb2mad = mad_b2.survey(**starting).dframe()
survb4xt = line_b4.survey(**starting).to_pandas(index="name")
survb2xt = line_b4.survey(**starting).reverse().to_pandas(index="name")

# ================================
plt.close('all')
plt.figure(1, figsize=(6.4*1.4, 4.8*1.2))

sp_order = [1, 3, 5, 2, 4, 6]

for ii, coordi in enumerate("X Y Z theta phi psi".split()):
    plt.subplot(3, 2, sp_order[ii])

    plt.plot(survb2mad["s"], survb2mad[coordi.lower()], "r", lw=5, alpha=0.5)
    plt.plot(survb2mad.loc["ip2", "s"], survb2mad.loc["ip2", coordi.lower()],
            marker="o", color="red", ms=5, alpha=0.5)
    plt.plot(survb2xt["s"], survb2xt[coordi], "-", color="darkred", lw=2)
    plt.plot(survb2xt.loc["ip2", "s"], survb2xt.loc["ip2", coordi],
            marker="x", color='darkred')

    plt.plot(survb4mad["s"], survb4mad[coordi.lower()], "b", lw=5, alpha=0.5)
    plt.plot(survb4mad.loc["ip2", "s"], survb4mad.loc["ip2", coordi.lower()],
            marker="o", color="blue", ms=5, alpha=0.5)
    plt.plot(survb4xt["s"], survb4xt[coordi], "-", color="darkblue", lw=2)
    plt.plot(survb4xt.loc["ip2", "s"], survb4xt.loc["ip2", coordi],
            marker="x", color='darkblue')

    plt.ylabel(coordi)
plt.subplots_adjust(left=.105, right=.95, wspace=.23)
plt.show()


import numpy as np
import xtrack as xt
pi = np.pi

# We create an environment
env = xt.Environment()

# We define variables that we will use for controlling the length and the
# integrated strength of the focusing quadrupoles
env['l.quad'] = 0.3
env['k1l.qf.1'] = 0
env['k1l.qf.2'] = 0

# Expressions can be associated to any beam element property, when creating the
# element. For example:
lbend = 3
line = env.new_line(components=[
    env.new('mqf.1', xt.Quadrupole, length=0.3, k1='k1l.qf.1 / l.quad'),
    env.new('d1.1',  xt.Drift, length=1),
    env.new('mb1.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.1',  xt.Drift, length=1),

    env.new('mqd.1', xt.Quadrupole, length=0.3, k1=0), # k1 will be set later
    env.new('d3.1',  xt.Drift, length=1),
    env.new('mb2.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.1',  xt.Drift, length=1),

    env.new('mqf.2', xt.Quadrupole, length=0.3, k1='k1l.qf.2 / l.quad'),
    env.new('d1.2',  xt.Drift, length=1),
    env.new('mb1.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.2',  xt.Drift, length=1),

    env.new('mqd.2', xt.Quadrupole, length=0.3, k1=0), # k1 will be set later
    env.new('d3.2',  xt.Drift, length=1),
    env.new('mb2.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.2',  xt.Drift, length=1),
])
line.particle_ref = xt.Particles(p0c=1.2e9, mass0=xt.PROTON_MASS_EV)

# Expressions can also be assigned after the creation of the line. For example, we
# can set the integrated strength of the defocusing quadrupoles (note that line[...]
# is equivalent to env[...]):
line['k1l.qd.1'] = -0.1
line['k1l.qd.2'] = -0.1
line['mqd.1'].k1 = 'k1l.qd.1 / l.quad'
line['mqd.2'].k1 = 'k1l.qd.2 / l.quad'

# When a variable is changed, the corresponding element property is automatically
# updated:
line.vars['k1l.qf.1'] = 0.1
line['mqf.1'].k1 # is 0.333, i.e. 0.1 / lquad

# Expressions can be modified after the creation of the line. For example, we can
# create a variable controlling the integrated strength of the two focusing quadrupoles
line['k1lf'] = 0.1
line['k1l.qf.1'] = 'k1lf'
line['k1l.qf.2'] = 'k1lf'
# and a variable controlling the integrated strength of the two defocusing quadrupoles
line['k1ld'] = -0.7
line['k1l.qd.1'] = 'k1ld'
line['k1l.qd.2'] = 'k1ld'

# Changes on the controlling variable are propagated to the two controlled ones and
# to the corresponding element properties:
line['k1lf'] = 0.2
line['k1l.qf.1'] # is 0.2
line['k1l.qf.2'] # is 0.2
line['mqf.1'].k1 # is 0.666, i.e. 0.2 / lquad
line['mqf.2'].k1 # is 0.666, i.e. 0.2 / lquad

# The `info()` method of a variable provides information on the existing relations
# between the variables. For example:
line.info('k1l.qf.1')
# prints:
##  vars['k1l.qf.1']._get_value()
#   vars['k1l.qf.1'] = 0.2
#
##  vars['k1l.qf.1']._expr
#   vars['k1l.qf.1'] = vars['k1lf']
#
##  vars['k1l.qf.1']._expr._get_dependencies()
#   vars['k1lf'] = 0.2
#
##  vars['k1l.qf.1']._find_dependant_targets()
#   element_refs['mqf.1'].k1

line.info('k1lf')
# prints:
##  vars['k1lf']._get_value()
#   vars['k1lf'] = 0.2
#
##  vars['k1lf']._expr is None
#
##  vars['k1lf']._find_dependant_targets()
#   vars['k1l.qf.2']
#   element_refs['mqf.2'].k1
#   vars['k1l.qf.1']
#   element_refs['mqf.1'].k1

# The `get_info()` method of an element provides information on attribute of
# an element. For example:
line['mqf.1'].get_info('k1')
# prints:
##  element_refs['mqf.1'].k1._get_value()
#   element_refs['mqf.1'].k1 = 0.6666666666666667
#
##  element_refs['mqf.1'].k1._expr
#   element_refs['mqf.1'].k1 = (vars['k1l.qf.1'] / 0.3)
#
##  element_refs['mqf.1'].k1._expr._get_dependencies()
#   vars['k1l.qf.1'] = 0.2
#
##  element_refs['mqf.1'].k1 does not influence any target

# Expressions can include multiple variables and mathematical operations. For example
line['a'] = '3 * sqrt(k1lf) + 2 * k1ld'

# The `line.vars.get_table()` method returns a table with the value of all the
# existing variables:
line.vars.get_table()
# returns:
#
# VarsTable: 9 rows, 3 cols
# name             value expr
# t_turn_s             0 None
# l.quad             0.3 None
# k1l.qf.1           0.2 k1lf
# k1l.qf.2           0.2 k1lf
# k1l.qd.1          -0.7 k1ld
# k1l.qd.2          -0.7 k1ld
# k1lf               0.2 None
# k1ld              -0.7 None
# a           -0.0583592 ((3.0 * sqrt(k1lf)) + (2.0 * k1ld))


# Regular expressions can be used to select variables. For example we can select all
# the variables containing `qf` using the following:
var_tab = line.vars.get_table()
var_tab.rows['.*qf.*']
# returns:
#
# VarsTable: 2 rows, 3 cols
# name             value expr
# k1l.qf.1           0.2 k1lf
# k1l.qf.2           0.2 k1lf

# xdeps is the library implementing the expression logic.
# xdeps has the concept of references and references can also be used to build
# expressions.
# References can be created with
env.ref['mqf.1'].k1 # -> element_refs['mqf.1'].k1
env.ref['k1lf'] # -> vars['k1lf']
# Expressions can be built with normal operators
env.ref['k1lf']*3+1 #-> ((vars['k1lf'] * 3) + 1)

# When assigning an expression to a quantity using the environment, e.g.
env['mqf.1'].k1 = env.ref['k1lf']*3+1
# the xdeps Manager that is inside the environment
env.ref_manager
# establishes a relation between lhs and rhs if the rhs is an expression or
# remove old relations and assign the value if the rhs is a value.
env['mqf.1'].get_info("k1")
# returns
# element_refs['mqf.1'].k1 = ((vars['k1lf'] * 3) + 1)

# It is also possible to assign to refs to obtain the same results:
env.ref['mqf.1'].k1 = env.ref['k1lf']*3+1

# The difference in between the two cases lies in the handling of the "+="
# operator. In Python
env['mqf.1'].k1+=1
# is equivalent to
env['mqf.1'].k1=env['mqf.1'].k1+1
# so the expression is deleted:
env['mqf.1'].get_info("k1")
# returns:
# element_refs['mqf.1'].k1 = 3.6
# element_refs['mqf.1'].k1._expr is None
# while
env.ref['mqf.1'].k1=env.ref['k1lf']*2
env.ref['mqf.1'].k1+=env.ref['k1ld']*3
# is equivalent to
# env.ref['mqf.1'].k1._expr + env.ref['k1ld']*3
# resulting in
# element_refs['mqf.1'].k1 = ((vars['k1lf'] * 2) + (vars['k1ld'] * 3))



import xtrack as xt
import xpart as xp
import xobjects as xo
import pathlib
import numpy as np

from cpymad.madx import Madx

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

def test_lhc_environment():

    mad1=Madx()
    mad1.call(str(test_data_folder / 'hllhc15_thick/lhc.seq'))
    mad1.call(str(test_data_folder / 'hllhc15_thick/hllhc_sequence.madx'))
    mad1.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
    mad1.use('lhcb1')
    mad1.call(str(test_data_folder / "hllhc15_thick/opt_round_150_1500.madx"))
    mad1.twiss()

    mad4=Madx()
    mad4.input('mylhcbeam=4')
    mad4.call(str(test_data_folder / 'hllhc15_thick/lhcb4.seq'))
    mad4.call(str(test_data_folder / 'hllhc15_thick/hllhc_sequence.madx'))
    mad4.input('beam, sequence=lhcb2, particle=proton, energy=7000;')
    mad4.use('lhcb2')
    mad4.call(str(test_data_folder / 'hllhc15_thick/opt_round_150_1500.madx'))
    mad4.twiss()


    line1=xt.Line.from_madx_sequence(mad1.sequence.lhcb1,
                                    allow_thick=True,
                                    deferred_expressions=True,
                                    replace_in_expr={'bv_aux':'bvaux_b1'})
    line1.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)

    line4=xt.Line.from_madx_sequence(mad4.sequence.lhcb2,
                                    allow_thick=True,
                                    deferred_expressions=True,
                                    replace_in_expr={'bv_aux':'bvaux_b2'})
    line4.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)
    line4.twiss_default['reverse'] = True

    env = xt.Environment(lines={'lhcb1': line1, 'lhcb2': line4})
    assert env.lhcb2.twiss_default['reverse'] == True

    env.to_json('lhc.json')

    env = xt.Environment.from_json('lhc.json')

    assert env.lhcb1.element_dict is env.element_dict
    assert env.lhcb2.element_dict is env.element_dict

    assert 'drift_1' in env.lhcb1.element_names
    assert 'drift_1' not in env.lhcb2.element_names

    assert 'ip1' in env.lhcb1.element_names
    assert 'ip1' in env.lhcb2.element_names
    assert 'ip1/b2' not in env.lhcb2.element_names

    assert 'mqxfa.a1r1/lhcb1' in env.lhcb1.element_names
    assert 'mqxfa.a1r1/lhcb2' in env.lhcb2.element_names
    assert 'mqxfa.a1r1'  not in env.lhcb1.element_names
    assert 'mqxfa.a1r1'  not in env.lhcb2.element_names

    assert 'mq.12l6.b1' in env.lhcb1.element_names
    assert 'mq.12l6.b2' in env.lhcb2.element_names
    assert 'mq.12l6.b2/lhcb2' not in env.lhcb2.element_names

    twb1 = env.lhcb1.twiss4d()
    twb2 = env.lhcb2.twiss4d()

    assert np.all(twb1.rows['ip.*'].name ==
        np.array(['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8', 'ip1.l1']))
    assert np.all(twb2.rows['ip.*'].name ==
        np.array(['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8', 'ip1.l1']))

    xo.assert_allclose(twb1.qx, 62.31, rtol=0, atol=1e-5)
    xo.assert_allclose(twb1.qy, 60.32, rtol=0, atol=1e-5)
    xo.assert_allclose(twb2.qx, 62.31, rtol=0, atol=1e-5)
    xo.assert_allclose(twb2.qy, 60.32, rtol=0, atol=1e-5)

    env['on_disp'] = 0
    env['on_x1'] = 10
    env['on_x5'] = 12

    twb1cross = env.lhcb1.twiss4d()
    twb2cross = env.lhcb2.twiss4d()

    xo.assert_allclose(twb1cross['px', 'ip1'], +10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twb2cross['px', 'ip1'], -10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twb1cross['py', 'ip5'], +12e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twb2cross['py', 'ip5'], -12e-6, rtol=0, atol=1e-9)

    # Check combined twiss
    twcross = env.twiss(method='4d')
    xo.assert_allclose(twcross.lhcb1['px', 'ip1'], +10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twcross.lhcb2['px', 'ip1'], -10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twcross.lhcb1['py', 'ip5'], +12e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twcross.lhcb2['py', 'ip5'], -12e-6, rtol=0, atol=1e-9)

    env.discard_trackers()
    assert env.lhcb1.tracker is None
    assert env.lhcb2.tracker is None

    mycontext = xo.ContextCpu()
    env.build_trackers(_context=mycontext)
    assert env.lhcb1.tracker is not None
    assert env.lhcb2.tracker is not None
    assert env.lhcb1.tracker._context is mycontext
    assert env.lhcb2.tracker._context is mycontext


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import matplotlib.pyplot as plt

# https://gitlab.cern.ch/e-beam/pybetacool
from pybetacool import PyBetaCool

# unofficial Betacool: https://github.com/dgamba/betacool
obj = PyBetaCool(BLDfilename='LEIR.bld', betacoolExec='/home/pkruyt/cernbox/BETACOOL_ecooling/betacool/Betacool_Linux')

# Execute Betacool emittance
obj.runBetacool('/m')

#################################
# parse file and plot data
#################################

tmp  = obj.parseCurveFile('emittance.cur')
time = tmp[:,0]
emittance = tmp[:,1]*1e-6

plt.figure()
plt.plot(time,emittance, label='emittance')
plt.title('betacool')
plt.ylabel('Emittance')
plt.xlabel('Time')
plt.legend()
plt.show()

#################################
# save data as reference for test
#################################
filpath='../emittance_betacool.npz'
np.savez(filpath, time=time, emittance=emittance)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import time
import numpy as np

import xtrack as xt
import xobjects as xo

num_turns = 100
num_particles = 1000

# Select a running mode
context = xo.ContextCpu()                       # Serial single core
# context = xo.ContextCpu(omp_num_threads=4)      # Serial multi core (4 threads)
# context = xo.ContextCpu(omp_num_threads='auto') # Serial multi core (automatic n. threads)

#################################
# Load a line and build tracker #
#################################

line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7e12)

line.build_tracker(_context=context, compile=False)

###########################
# Generate some particles #
###########################

particles = line.build_particles(
    x_norm=np.linspace(-2, 2, num_particles), y_norm=0.1, delta=3e-4,
    nemitt_x=2.5e-6, nemitt_y=2.5e-6)

####################
# Optimize tracker #
####################

line.optimize_for_tracking()

###########################
# Track with optimization #
###########################

print('Start tracking')
t1 = time.time()
line.track(particles, num_turns=num_turns, time=True)
tracking_time = line.time_last_track
t2=time.time()

particles.move(_context = xo.ContextCpu())

assert len(particles.state) == num_particles
assert np.all(particles.state == 1)

print(f'Tracking time [us/p/turn]: {tracking_time*1e6/num_particles/num_turns:.2f}')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time

import numpy as np
from matplotlib import pyplot as plt
from scipy import constants

import xobjects as xo
context = xo.ContextCpu(omp_num_threads=0)
import xtrack as xt
import xfields as xf
import xpart as xp

n_macroparticles = int(1e3)
nemitt_x = 2E-6
nemitt_y = 2E-6
p0c = 7e12
gamma = p0c/xp.PROTON_MASS_EV
physemit_x = nemitt_x/gamma
physemit_y = nemitt_y/gamma
beta_x = 1.0
beta_y = 1.0

sigma_z = 0.08
momentumCompaction = 3.483575072011584e-04
eta = momentumCompaction - 1.0 / gamma ** 2
circumference = 26658.883199999
frev = constants.c/circumference
f_RF = 400.8E6
h = f_RF / frev
betar = np.sqrt(1 - 1 / gamma ** 2)
p0 = constants.m_p * betar * gamma * constants.c
voltage = 16E6
Qs = np.sqrt(constants.e * voltage * eta * h / (2 * np.pi * betar * constants.c * p0))
beta_s = eta * circumference / (2 * np.pi * Qs);
sigma_delta = sigma_z / beta_s
Qx = 0.31
Qy = 0.32
bucket_length = 2.5E-9

linear_fixed_rf = True

particles = xp.Particles(_context=context,
    p0c=p0c,
    x=np.zeros(n_macroparticles),
    px=np.zeros(n_macroparticles),
    y=np.zeros(n_macroparticles),
    py=np.zeros(n_macroparticles),
    zeta=bucket_length*constants.c*betar*np.linspace(-5,5,n_macroparticles),
    delta=np.zeros(n_macroparticles),
)

if linear_fixed_rf:
    arc = xt.LineSegmentMap(
            betx=beta_x,qx=Qx,
            bety=beta_y,qy=Qy,
            longitudinal_mode='linear_fixed_rf',
            voltage_rf = voltage*10,
            frequency_rf = f_RF/10,
            lag_rf = 180.0,
            slippage_length = circumference,
            momentum_compaction_factor = momentumCompaction)
else:
    arc = xt.LineSegmentMap(
            betx=beta_x,qx=Qx,
            bety=beta_y,qy=Qy,
            bets=-beta_s,qs=Qs)

line = xt.Line(elements=[arc])
line.build_tracker()
line.track(particles,num_turns=int(1/Qs),turn_by_turn_monitor=True)
mon = line.record_last_track

plt.figure()
plt.plot(mon.zeta/(bucket_length*constants.c*betar),mon.delta/sigma_delta,'.r')

particles = xp.Particles(_context=context,
    p0c=p0c,
    x=np.zeros(n_macroparticles),
    px=np.zeros(n_macroparticles),
    y=np.zeros(n_macroparticles),
    py=np.zeros(n_macroparticles),
    zeta=bucket_length*constants.c*betar*np.linspace(-5,5,n_macroparticles),
    delta=np.zeros(n_macroparticles),
)

if linear_fixed_rf:
    arc = xt.LineSegmentMap(
            betx=beta_x,qx=Qx,
            bety=beta_y,qy=Qy,
            longitudinal_mode='linear_fixed_rf',
            voltage_rf = voltage,
            frequency_rf = f_RF,
            lag_rf = 180.0,
            slippage_length = circumference,
            momentum_compaction_factor = momentumCompaction)
else:
    arc = xt.LineSegmentMap(
            betx=beta_x,qx=Qx,
            bety=beta_y,qy=Qy,
            bets=-beta_s,qs=Qs,bucket_length=bucket_length)

line2 = xt.Line(elements=[arc])
line2.build_tracker()
line2.track(particles,num_turns=int(1/Qs),turn_by_turn_monitor=True)
mon = line2.record_last_track

plt.plot(mon.zeta/(bucket_length*constants.c*betar),mon.delta/sigma_delta,'.b')
plt.xlabel(r'$z$ [bucket_lenght$\cdot c\beta_r$]')
plt.ylabel('$\delta$ [$\sigma_{\delta}$]')
plt.show()





import numpy as np
import xtrack as xt

opt_from_callable = xt.match.opt_from_callable

def my_function(x):
    return [(x[0]-0.0001)**2, (x[1]-0.0003)**2, (x[2]+0.0005)**2]

x0 = [0., 0., 0.]

opt = opt_from_callable(my_function, x0=x0, steps=[1e-6, 1e-6, 1e-6],
                        tar=[0., 0., 0.], tols=[1e-12, 1e-12, 1e-12])
opt.solve()




import numpy as np

import xtrack as xt
import xobjects as xo

env = xt.Environment()

# A simple line made of quadrupoles spaced by 5 m
env.new_line(name='l1', components=[
    env.new('q1', 'Quadrupole', length=2.0, at=0., anchor='start'),
    env.new('q2', 'Quadrupole', length=2.0, anchor='start', at=5., from_='q1@end'),
    env.new('q3', 'Quadrupole', length=2.0, anchor='start', at=5., from_='q2@end'),
    env.new('q4', 'Quadrupole', length=2.0, anchor='start', at=5., from_='q3@end'),
    env.new('q5', 'Quadrupole', length=2.0, anchor='start', at=5., from_='q4@end'),
])

# Test absolute anchor of start 'l1'
env.new_line(name='lstart', components=[
    env.place('l1', anchor='start', at=10.),
])

# Test absolute anchor of end 'l1'
env.new_line(name='lend', components=[
    env.place('l1', anchor='end', at=40.),
])

# Test absolute anchor of center 'l1'
env.new_line(name='lcenter', components=[
    env.place('l1', anchor='center', at=25.),
])

# Test relative anchor of start 'l1' to start of another element
env.new_line(name='lstcnt', components=[
    env.new('q0', 'Quadrupole', length=2.0, at=5.),
    env.place('l1', anchor='start', at=5., from_='q0@center'),
])

# Test relative anchor of start 'l1' to end of another element
env.new_line(name='lstst', components=[
    env.place('q0', at=5.),
    env.place('l1', anchor='start', at=5. + 1., from_='q0@end'),
])


# Test relative anchor of start 'l1' to end of another element
env.new_line(name='lstend', components=[
    env.place('q0', at=5.),
    env.place('l1', anchor='start', at=5. - 1., from_='q0@end'),
])

tt_l1 = env['l1'].get_table()
tt_test = tt_l1
assert np.all(tt_test.name == np.array(
    ['q1', 'drift_1', 'q2', 'drift_2', 'q3', 'drift_3', 'q4', 'drift_4',
     'q5', '_end_point']))
xo.assert_allclose(tt_test.s, np.array(
    [ 0.,  2.,  7.,  9., 14., 16., 21., 23., 28., 30.]),
    rtol=0., atol=1e-15)
xo.assert_allclose(tt_test.s_start, np.array(
    [ 0.,  2.,  7.,  9., 14., 16., 21., 23., 28., 30.]),
    rtol=0., atol=1e-15)
xo.assert_allclose(tt_test.s_center, np.array(
    [ 1. ,  4.5,  8. , 11.5, 15. , 18.5, 22. , 25.5, 29. , 30. ]),
    rtol=0., atol=1e-15)
xo.assert_allclose(tt_test.s_end, np.array(
    [ 2.,  7.,  9., 14., 16., 21., 23., 28., 30., 30.]),
    rtol=0., atol=1e-15)

tt_lstart = env['lstart'].get_table()
tt_test = tt_lstart
assert np.all(tt_test.name == np.array(
    ['drift_5', 'q1', 'drift_1', 'q2', 'drift_2', 'q3', 'drift_3', 'q4',
       'drift_4', 'q5', '_end_point']))
xo.assert_allclose(tt_test.s, np.array(
    [ 0., 10., 12., 17., 19., 24., 26., 31., 33., 38., 40.]),
    rtol=0., atol=1e-15)

tt_lend = env['lend'].get_table()
tt_test = tt_lend
assert np.all(tt_test.name == np.array(
    ['drift_6', 'q1', 'drift_1', 'q2', 'drift_2', 'q3', 'drift_3', 'q4',
    'drift_4', 'q5', '_end_point']))
xo.assert_allclose(tt_test.s, np.array(
    [ 0., 10., 12., 17., 19., 24., 26., 31., 33., 38., 40.]),
    rtol=0., atol=1e-15)

tt_lcenter = env['lcenter'].get_table()
tt_test = tt_lcenter
assert np.all(tt_test.name == np.array(
    ['drift_7', 'q1', 'drift_1', 'q2', 'drift_2', 'q3', 'drift_3', 'q4',
    'drift_4', 'q5', '_end_point']))
xo.assert_allclose(tt_test.s, np.array(
    [ 0., 10., 12., 17., 19., 24., 26., 31., 33., 38., 40.]),
    rtol=0., atol=1e-15)

tt_lstcnt = env['lstcnt'].get_table()
tt_test = tt_lstcnt
assert np.all(tt_test.name == np.array(
    ['drift_8', 'q0', 'drift_9', 'q1', 'drift_1', 'q2', 'drift_2', 'q3',
     'drift_3', 'q4', 'drift_4', 'q5', '_end_point']))
xo.assert_allclose(tt_test.s, np.array(
    [ 0.,  4.,  6., 10., 12., 17., 19., 24., 26., 31., 33., 38., 40.]),
    rtol=0., atol=1e-15)

tt_lstst = env['lstst'].get_table()
tt_test = tt_lstst
assert np.all(tt_test.name == np.array(
    ['drift_10', 'q0', 'drift_11', 'q1', 'drift_1', 'q2', 'drift_2', 'q3',
     'drift_3', 'q4', 'drift_4', 'q5', '_end_point']))

tt_lstend = env['lstend'].get_table()
tt_test = tt_lstend
assert np.all(tt_test.name == np.array(
    ['drift_12', 'q0', 'drift_13', 'q1', 'drift_1', 'q2', 'drift_2', 'q3',
     'drift_3', 'q4', 'drift_4', 'q5', '_end_point']))



import numpy as np
import xtrack as xt
import xobjects as xo

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')

tt = line.get_table()
tt_quads = tt.rows[tt.element_type=='Quadrupole']

# Introduce misalignments on all quadrupoles
tt = line.get_table()
tt_quad = tt.rows['mq\..*']
rgen = np.random.RandomState(1) # fix seed for random number generator
                                # (to have reproducible results)
shift_x = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
shift_y = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
rot_s = rgen.randn(len(tt_quad)) * 1e-3 # 1 mrad rms rotation on all quads
k2l = rgen.rand(len(tt_quad)) * 1e-3

line['on_error'] = 1.
for nn_quad, sx, sy, rr, kkk in zip(tt_quad.name, shift_x, shift_y, rot_s, k2l):
    line[nn_quad].shift_x = sx * line.ref['on_error']
    line[nn_quad].shift_y = sy * line.ref['on_error']
    line[nn_quad].rot_s_rad = rr * line.ref['on_error']
    line[nn_quad].knl[2] = kkk * line.ref['on_error']
tw = line.madng_twiss()

xo.assert_allclose(tw.x, tw.x_ng, atol=5e-4*tw.x.std(), rtol=0)
xo.assert_allclose(tw.y, tw.y_ng, atol=5e-4*tw.y.std(), rtol=0)
xo.assert_allclose(tw.betx2, tw.beta12_ng, atol=0, rtol=2e-3)
xo.assert_allclose(tw.bety1, tw.beta21_ng, atol=0, rtol=2e-3)
xo.assert_allclose(tw.wx_chrom, tw.wx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
xo.assert_allclose(tw.wy_chrom, tw.wy_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
xo.assert_allclose(tw.ax_chrom, tw.ax_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
xo.assert_allclose(tw.ay_chrom, tw.ay_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
xo.assert_allclose(tw.bx_chrom, tw.bx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
xo.assert_allclose(tw.by_chrom, tw.by_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)

line['on_error'] = 0
tw = line.madng_twiss()
xo.assert_allclose(tw.x, 0, atol=1e-10, rtol=0)
xo.assert_allclose(tw.y, 0, atol=1e-10, rtol=0)
xo.assert_allclose(tw.betx2, 0, atol=1e-10, rtol=0)
xo.assert_allclose(tw.bety1, 0, atol=1e-10, rtol=0)
xo.assert_allclose(tw.x, tw.x_ng, atol=1e-10, rtol=0)
xo.assert_allclose(tw.y, tw.y_ng, atol=1e-10, rtol=0)
xo.assert_allclose(tw.betx2, tw.beta12_ng, atol=1e-10, rtol=0)
xo.assert_allclose(tw.bety1, tw.beta21_ng, atol=1e-19, rtol=0)
xo.assert_allclose(tw.wx_chrom, tw.wx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
xo.assert_allclose(tw.wy_chrom, tw.wy_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
xo.assert_allclose(tw.ax_chrom, tw.ax_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
xo.assert_allclose(tw.ay_chrom, tw.ay_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
xo.assert_allclose(tw.bx_chrom, tw.bx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
xo.assert_allclose(tw.by_chrom, tw.by_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)


import xtrack as xt
import numpy as np

assert_allclose = np.testing.assert_allclose

k0 = 2.
length = 0.1
rot_s_rad = 0.2

bend = xt.Bend(k0=k0, length=length, rot_s_rad=rot_s_rad)

assert bend.shift_x == 0
assert bend.shift_y == 0
assert_allclose(bend._sin_rot_s, np.sin(rot_s_rad), rtol=0, atol=1e-14)
assert_allclose(bend._cos_rot_s, np.cos(rot_s_rad), rtol=0, atol=1e-14)

p = xt.Particles(x=0, p0c=1e12)
bend.track(p)

assert_allclose = np.testing.assert_allclose
assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3, atol=0)
assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3, atol=0)


# Change the shift
rot_s_rad = 0.3
bend.rot_s_rad = rot_s_rad
p = xt.Particles(x=0, p0c=1e12)
bend.track(p)
assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3, atol=0)
assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3, atol=0)

# Make a line
line = xt.Line(elements=[bend])

# Slice the line:
line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Uniform(3))])
line.build_tracker()

tt = line.get_table()
assert len(tt.rows['e0\.\..?']) == 3

p = xt.Particles(x=0, p0c=1e12)
line.track(p)

assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3, atol=0)
assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3, atol=0)

# Change the shift
rot_s_rad = 0.4
bend.rot_s_rad = rot_s_rad
p = xt.Particles(x=0, p0c=1e12)
line.track(p)

assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3, atol=0)
assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3, atol=0)


import xobjects as xo

COAST_STATE_RANGE_START= 1000000
DEFAULT_FRAME_RELATIVE_LENGTH = 0.9

class SyncTime:

    def __init__(self, circumference, id, frame_relative_length=None,
                 at_start=False, at_end=False):
        if frame_relative_length is None:
            frame_relative_length = DEFAULT_FRAME_RELATIVE_LENGTH
        assert id > COAST_STATE_RANGE_START
        self.id = id
        self.frame_relative_length = frame_relative_length
        self.circumference = circumference
        self.at_start = at_start
        self.at_end = at_end

    def track(self, particles):

        assert isinstance(particles._context, xo.ContextCpu), (
                          'SyncTime only available for CPU for now')

        beta0 = particles._xobject.beta0[0]
        beta1 = beta0 / self.frame_relative_length
        beta0_beta1 = beta0 / beta1

        mask_alive = particles.state > 0

        zeta_min = -self.circumference/ 2 * beta0_beta1 + particles.s * (
                   1 - beta0_beta1)

        if (self.at_start and particles.at_turn[0] == 0
                and not (particles.state == -COAST_STATE_RANGE_START).any()): # done by the user
            mask_stop = mask_alive * (particles.zeta < zeta_min)
            particles.state[mask_stop] = -COAST_STATE_RANGE_START
            particles.zeta[mask_stop] += self.circumference * beta0 / beta1

        # Resume particles previously stopped
        particles.state[particles.state==-self.id] = 1
        particles.reorganize()

        # Identify particles that need to be stopped
        zeta_min = -self.circumference/ 2 * beta0_beta1 + particles.s * (1 - beta0_beta1)
        mask_stop = mask_alive & (particles.zeta < zeta_min)

        # Check if some particles are too fast
        mask_too_fast = mask_alive & (
            particles.zeta > zeta_min + self.circumference * beta0_beta1)
        if mask_too_fast.any():
            raise ValueError('Some particles move faster than the time window')

        # For debugging (expected to be triggered)
        # mask_out_of_circumference = mask_alive & (
        #       (particles.zeta > self.circumference / 2)
        #     | (particles.zeta < -self.circumference / 2))
        # if mask_out_of_circumference.any():
        #     raise ValueError('Some particles are out of the circumference')

        # Update zeta for particles that are stopped
        particles.zeta[mask_stop] += beta0_beta1 * self.circumference

        # Stop particles
        particles.state[mask_stop] = -self.id

        if self.at_end:
            mask_alive = particles.state > 0
            particles.zeta[mask_alive] -= (
                self.circumference * (1 - beta0_beta1))

        if self.at_end and particles.at_turn[0] == 0:
            particles.state[particles.state==-COAST_STATE_RANGE_START] = 1

def install_sync_time_at_collective_elements(line, frame_relative_length=None):

    circumference = line.get_length()

    ltab = line.get_table()
    tab_collective = ltab.rows[ltab.iscollective]
    for ii, nn in enumerate(tab_collective.name):
        cc = x=SyncTime(circumference=circumference,
                        frame_relative_length=frame_relative_length,
                        id=COAST_STATE_RANGE_START + ii + 1)
        line.insert_element(element=cc, name=f'synctime_{ii}', at=nn)

    synctime_start = SyncTime(circumference=circumference,
                        frame_relative_length=frame_relative_length,
                        id=COAST_STATE_RANGE_START + len(tab_collective)+1,
                        at_start=True)
    synctime_end = SyncTime(circumference=circumference,
                        frame_relative_length=frame_relative_length,
                        id=COAST_STATE_RANGE_START + len(tab_collective)+2,
                        at_end=True)

    line.insert_element(element=synctime_start, name='synctime_start', at_s=0)
    line.append_element(synctime_end, name='synctime_end')

def prepare_particles_for_sync_time(particles, line):
    synctime_start = line['synctime_start']
    beta0 = particles._xobject.beta0[0]
    beta1 = beta0 / synctime_start.frame_relative_length
    beta0_beta1 = beta0 / beta1
    zeta_min = -synctime_start.circumference/ 2 * beta0_beta1 + particles.s * (
                1 - beta0_beta1)
    mask_alive = particles.state > 0
    mask_stop = mask_alive * (particles.zeta < zeta_min)
    particles.state[mask_stop] = -COAST_STATE_RANGE_START
    particles.zeta[mask_stop] += synctime_start.circumference * beta0 / beta1


import xtrack as xt



fun_container = xt.line.Functions()
_functions = []
for ff in fun_container._mathfunctions.keys():
    _functions.append(ff)

def get_mad_str_expr(var_expr):
    str_expr = str(var_expr)
    mad_str_expr = str_expr.replace("vars['", '').replace("']", '')
    for ff in _functions:
        mad_str_expr = mad_str_expr.replace('f.' + ff, ff)
    return mad_str_expr

def extract_val_or_madexpr(var, dct_expr, dct_val):
    var_name = var._key
    if var_name in dct_expr or var_name in dct_val:
        return
    if var_name in _functions:
        return
    if var._expr is not None:
        dct_expr[var_name] = get_mad_str_expr(var._expr)
        for vv in var._expr._get_dependencies():
            extract_val_or_madexpr(vv, dct_expr, dct_val)
        for vv in var._find_dependant_targets():
            if str(vv).startswith('vars['):
                extract_val_or_madexpr(vv, dct_expr, dct_val)
    else:
        dct_val[var_name] = var._value


def gen_madx_optics_file_auto(collider, fname):

    dct_expr = {}
    dct_val = {}

    vtable = collider.vars.get_table()
    vsave = vtable.rows["(acb.*)|(kd.*)|(kq.*)|(ks.*)"]
    for nn in vsave.name:
        nn = str(nn) # to avoid numpy string
        vv = collider.vars[nn]
        extract_val_or_madexpr(vv, dct_expr, dct_val)

    out_lines = []

    for nn in sorted(dct_val.keys()):
        out_lines.append(nn + ' = ' + str(dct_val[nn]) + ';')

    out_lines.append('')

    for nn in sorted(dct_expr.keys()):
        out_lines.append(nn + ' := ' + dct_expr[nn] + ';')

    with open(fname, 'w') as fid:
        fid.write('\n'.join(out_lines))



import matplotlib.pyplot as plt
from pathlib import Path
from time import time

import xtrack as xt
import numpy as np

all_num_cuts = np.logspace(np.log2(10), np.log2(50_000), base=2, num=15)
time_to_cut = []

line0 = xt.Line.from_json(
    Path(__file__).parent /
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json'
)

for num_cuts in all_num_cuts:
    ss = np.linspace(0, line0.get_length(), num=int(num_cuts))
    line = line0.copy()
    start = time()
    line.cut_at_s(s=list(ss))
    end = time()
    time_to_cut.append(end - start)

fig, ax = plt.subplots()
ax.plot(all_num_cuts, time_to_cut)
ax.set_title("Slicing time vs number of cuts")
ax.set_ylabel('time [s]')
ax.set_xlabel('number of cuts [1]')
plt.show()



import xtrack as xt
import xobjects as xo
import numpy as np

from xtrack.beam_elements.magnets import MagnetEdge

def compare(e_test, e_ref):
    p0 = xt.Particles(kinetic_energy0=50e6,
                      x=1e-2, y=2e-2, zeta=1e-2, px=10e-2, py=20e-2, delta=1e-2)

    e_test.compile_kernels()

    # Expanded drift
    p_test = p0.copy()
    p_ref = p0.copy()

    e_test.track(p_test)

    if isinstance(e_ref, list):
        mini_line = xt.Line(elements=e_ref)
        mini_line.build_tracker()
        mini_line.track(p_ref)
    else:
        e_ref.track(p_ref)

    xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)


print('==> Supressed')
e_test = MagnetEdge(model='suppressed', kn=[0], ks=[0])
e_ref = xt.DipoleEdge(model='suppressed', k=0)
compare(e_test, e_ref)

print('==> Linear but does nothing')
e_test = MagnetEdge(model='linear', kn=[0], ks=[0])
e_ref = xt.DipoleEdge(model='linear', k=0)
compare(e_test, e_ref)

print('==> Full but does nothing')
e_test = MagnetEdge(model='full', kn=[0], ks=[0])
e_ref = xt.DipoleEdge(model='full', k=0)
compare(e_test, e_ref)

print('==> Only linear')
e_test = MagnetEdge(model='linear', kn=[3], face_angle=0.1, face_angle_feed_down=0.2, fringe_integral=0.3, half_gap=0.4)
e_ref = xt.DipoleEdge(model='linear', k=3, e1=0.1, e1_fd=0.2, fint=0.3, hgap=0.4)
compare(e_test, e_ref)

print('==> Full but with only the dipole component')
e_test = MagnetEdge(model='full', kn=[3], face_angle=0.1, face_angle_feed_down=0.2, fringe_integral=0.3, half_gap=0.4)
e_ref = xt.DipoleEdge(model='full', k=3, e1=0.1, e1_fd=0.2, fint=0.3, hgap=0.4)
compare(e_test, e_ref)

print('==> Multipole fringe, without the dipole component')
e_test = MagnetEdge(model='full', kn=[0, 2, 3], k_order=2)
e_ref = xt.MultipoleEdge(kn=[0, 2, 3], order=2)
compare(e_test, e_ref)

print('==> Full model with a dipole component, no angle')
e_test = MagnetEdge(model='full', kn=[3, 4, 5], fringe_integral=0.3, half_gap=0.4, k_order=2)
sandwich = [
    xt.DipoleEdge(model='full', k=3, fint=0.3, hgap=0.4),
    xt.MultipoleEdge(kn=[0, 4, 5], order=2),
]
compare(e_test, sandwich)

print('==> Full model with a dipole component and angle')
e_test = MagnetEdge(model='full', kn=[3, 4, 5], face_angle=0.2, face_angle_feed_down=0.0, fringe_integral=0.3, half_gap=0.4, k_order=2)
sandwich = [
    xt.YRotation(angle=np.rad2deg(-0.2)),  # The rotation is also the other way than in the underlying map :'(
    xt.DipoleEdge(model='full', k=3, fint=0.3, hgap=0.4),
    xt.MultipoleEdge(kn=[0, 4, 5], order=2),
    xt.Wedge(angle=-0.2, k=3),
]
compare(e_test, sandwich)

print('==> Full model with a dipole component and angle, at the exit')
e_test = MagnetEdge(model='full', kn=[3, 4, 5], is_exit=True, face_angle=0.2, face_angle_feed_down=0.0, fringe_integral=0.3, half_gap=0.4, k_order=2)
sandwich = [
    xt.Wedge(angle=-0.2, k=3),
    xt.MultipoleEdge(kn=[0, 4, 5], is_exit=True, order=2),
    xt.DipoleEdge(model='full', k=-3, fint=0.3, hgap=0.4),
    xt.YRotation(angle=np.rad2deg(-0.2)),
]
compare(e_test, sandwich)

print('==> Linear at exit')
e_test = MagnetEdge(model='linear', is_exit=True, kn=[3], face_angle=0.1, face_angle_feed_down=0.2, fringe_integral=0.3, half_gap=0.4)
e_ref = xt.DipoleEdge(model='linear', k=3, e1=0.1, e1_fd=0.2, fint=0.3, hgap=0.4)
compare(e_test, e_ref)


"""
Exciter element

Author: Philipp Niedermayer
Date: 11.11.2022

"""


import xobjects as xo
import numpy as np

from ..base_element import BeamElement
from ..general import _pkg_root


class Exciter(BeamElement):
    """Beam element modeling a transverse exciter as a time-dependent thin multipole.

    The given multipole components (knl and ksl) are scaled according to a custom waveform,
    allowing for arbitrary time dependence. The waveform is specified by an array of samples:

        knl(t) = knl * samples[i]

    It is *not* assumed that the variations are slow compared to the revolution frequency
    and the particle arrival time is taken into account when determining the sample index:

        i = sampling_frequency * ( ( at_turn - start_turn ) / f_rev - zeta / beta0 / c0 )

    where zeta=(s-beta0*c0*t) is the longitudinal coordinate of the particle, beta0 the
    relativistic beta factor of the particle, c0 is the speed of light, at_turn is the
    current turn number, f_rev is the revolution frequency, and sampling_frequency is the sampling
    frequency. The excitation starts with the first sample when the reference particle 
    arrives at the element in start_turn.

    For example, to compute samples for a sinusoidal excitation with frequency f_ex one
    would calculate the waveform as: samples[i] = np.sin(2*np.pi*f_ex*i/sampling_frequency)

    Notes:
        - This is not to be confused with an RFMultipole, which inherits the characteristics
          of an RFCavity and whose oscillation is therefore with respect to the reference
          particle. While the frequency of the RFMultipole is therefore restricted to
          harmonics of the revolution frequency, the exciter allows for arbitrary frequencies.
        - This is also not to be confused with an ACDipole, for which the oscillation is 
          assumed to be slow compared to the revolution frequency and the kick is the same 
          for all particles independent of their longitudinal coordinate.


    Parameters:
        - knl (float array): Normalized integrated strength of the normal components. Unit: m^-n (n=0,1,2,...).
        - ksl (float array): Normalized integrated strength of the skew components. Unit: m^-n (n=0,1,2,...).
        - order (int): Multipole order (readonly), i.e. largest n with non-zero knl or ksl.
        - samples (float array): Samples of excitation strength to scale knl and ksl as function of time.
        - nsamples (int): Number of samples. Pass this instead of samples to reserve memory for later initialisation.
        - sampling_frequency (float): Sampling frequency in Hz.
        - frev (float): Revolution frequency in Hz of circulating beam (used to relate turn number to sample index).
        - start_turn (int): Turn of the reference particle when to start excitation.
        - duration (float): Duration of excitation in s (defaults to nsamples/sampling_frequency). Repeats the waveform to fill the duration.

    Example:
        >>> fs = 10e6 # sampling frequency in Hz
        >>> 
        >>> # load waveform into memory
        >>> signal = np.copy(np.memmap("signal.10MHz.float32", np.float32))
        >>> 
        >>> # alternatively compute samples on the fly, for example a simple sine at 500 kHz ...
        >>> t = np.arange(1000)/fs
        >>> f_ex = 5e5 # excitation frequency in Hz
        >>> signal = np.sin(2*np.pi*f_ex*t)
        >>> 
        >>> # ... or a sweep from 500 to 800 kHz
        >>> f_ex_1 = 8e5
        >>> signal = scipy.signal.chirp(t, f_ex, t[-1], f_ex_1)
        >>> 
        >>> # create the exciter
        >>> frev = 1e6 # revolution frequency in Hz
        >>> k0l = 0.1 # this is scaled by the waveform
        >>> exciter = Exciter(samples=signal, sampling_frequency=fs, frev=frev, start_turn=0, knl=[k0l])
        >>> 
        >>> # add it to the line
        >>> line.insert_element(index=..., name=..., element=exciter)

    """

    _xofields={
        'order': xo.Int64,
        'nsamples': xo.Int64,
        'sampling_frequency': xo.Float64,
        'frev': xo.Float64,
        'start_turn': xo.Int64,
        'nduration': xo.Int64,
        'knl': xo.Float64[:],
        'ksl': xo.Float64[:],
        'samples': xo.Float32[:],
        }

    has_backtrack = True

    _extra_c_sources = [_pkg_root.joinpath('beam_elements/elements_src/exciter.h')]


    def __init__(self, *, samples=None, nsamples=None, sampling_frequency=0, frev=0, knl=None, ksl=None, start_turn=0, duration=None, _xobject=None, **kwargs):

        if _xobject is not None:
            super().__init__(_xobject=_xobject)

        else:
            # The default is needed, as we wish to be able to instantiate
            # elements without properties (empty elements can be templates).
            if knl is None and ksl is None:
                knl = [0]
                ksl = []
            elif knl is None:
                knl = []
            elif ksl is None:
                ksl = []

            # sanitize knl and ksl array length
            n = max(len(knl), len(ksl))
            nknl = np.zeros(n, dtype=np.float64)
            nksl = np.zeros(n, dtype=np.float64)
            if knl is not None:
                nknl[:len(knl)] = np.array(knl)
            if ksl is not None:
                nksl[:len(ksl)] = np.array(ksl)
            kwargs["order"] = n - 1

            # determine sample length or create empty samples for later initialisation
            if samples is not None:
                if nsamples is not None and nsamples != len(samples):
                    raise ValueError("Only one of samples or nsamples may be specified")
                nsamples = len(samples)
            if samples is None:
                samples = np.zeros(nsamples)
            kwargs["nduration"] = nsamples if duration is None else (duration * sampling_frequency)

            super().__init__(
                knl=nknl, ksl=nksl, samples=samples, nsamples=nsamples,
                sampling_frequency=sampling_frequency,
                frev=frev, start_turn=start_turn, **kwargs)

    @property
    def duration(self):
        return self.nduration / self.sampling_frequency

    @duration.setter
    def duration(self, duration):
        self.nduration = duration * self.sampling_frequency

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        ctx2np = self._buffer.context.nparray_from_context_array
        return self.__class__(knl=-ctx2np(self.knl),
                              ksl=-ctx2np(self.ksl),
                              samples=self.samples,
                              nsamples=self.nsamples,
                              sampling_frequency=self.sampling_frequency,
                              frev=self.frev,
                              start_turn=self.start_turn,
                              _context=_context, _buffer=_buffer, _offset=_offset)



from pymadng import MAD
import matplotlib.pyplot as plt

from cpymad.madx import Madx
madx = Madx()
madx.call("../../test_data/sps_thick/sps.seq")
madx.call("../../test_data/sps_thick/lhc_q20.str")
madx.beam()
madx.use("sps")
twmadx = madx.twiss()

mad = MAD()

# with MAD() as mad:
if True:
    mad.MADX.load('"../../test_data/sps_thick/sps.seq"', "'sps.mad'")
    mad.MADX.load(f"'../../test_data/sps_thick/lhc_q20.str'", f"'lhc_q20.mad'")
    mad["sps"] = mad.MADX.sps
    mad.sps.beam = mad.beam()
    mad["mytwtable", 'mytwflow'] = mad.twiss(sequence=mad.sps, method=4, mapdef=2, implicit=True, nslice=3, save="'atbody'")

    dq1 = mad.mytwtable.dq1
    # plt.plot(mad.mytwtable.s, mad.mytwtable.beta11)

    mad.send('''
        local track in MAD  -- like "from MAD import track"
        local mytrktable, mytrkflow = MAD.track{sequence=MADX.sps, method=4,
                                                mapdef=4, nslice=3}

         -- print(MAD.typeid.is_damap(mytrkflow[1]))

        local normal in MAD.gphys  -- like "from MAD.gphys import normal"
        local my_norm_for = normal(mytrkflow[1]):analyse('anh') -- anh stands for anharmonicity

        local nf = my_norm_for
        py:send({
                nf:q1{1}, -- qx from the normal form (fractional part)
                nf:q2{1}, -- qy
                nf:dq1{1}, -- dqx / d delta
                nf:dq2{1}, -- dqy / d delta
                nf:dq1{2}, -- d2 qx / d delta2
                nf:dq2{2}, -- d2 qy / d delta2
                nf:anhx{1, 0}, -- dqx / djx
                nf:anhy{0, 1}, -- dqy / djy
                nf:anhx{0, 1}, -- dqx / djy
                nf:anhy{1, 0}, -- dqy / djx
             })
    ''')

    out = mad.recv()

    # mad["mytrktable", "mytrkflow"] = mad.track(sequence=mad.sps, method=4)
    # mad.gphys.normal(mad["mytrkflow"][0])



# plt.show()

# doc
# mad.show(mad.twiss)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

from cpymad.madx import Madx

import xpart as xp
import xobjects as xo
import xtrack as xt
import xfields as xf


seq_name = 'sps'
bunch_intensity = 1e11/3
sigma_z = 22.5e-2/3 # Short bunch to avoid probing bucket non-linearity
                    # to compare against frozen
nemitt_x=2.5e-6
nemitt_y=2.5e-6

mad = Madx()
mad.call('sps_thin.seq')
mad.use(seq_name)

line = xt.Line.from_madx_sequence(
                                            mad.sequence[seq_name],
                                            install_apertures=True)
# enable RF
V_RF = 3e6
line['acta.31637'].voltage = V_RF
line['acta.31637'].lag = 180.

# A test particle
part = xp.Particles(gamma0=mad.sequence[seq_name].beam.gamma,
                    mass0=xp.PROTON_MASS_EV)
line.particle_ref = part.copy()
part.x += 2e-3
part.y += 3e-3
part.zeta += 20e-2

# old-style output
with open('line_no_spacecharge_and_particle.json', 'w') as fid:
    json.dump({
        'line': line.to_dict(),
        'particle': part.to_dict()},
        fid, cls=xo.JEncoder, indent=4)

# new-style output
line.to_json('line_no_spacecharge.json')

lprofile = xf.LongitudinalProfileQGaussian(
        number_of_particles=bunch_intensity,
        sigma_z=sigma_z,
        z0=0.,
        q_parameter=1.)

xf.install_spacecharge_frozen(
            line=line,
            particle_ref=part,
            longitudinal_profile=lprofile,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y,
            sigma_z=sigma_z,
            num_spacecharge_interactions=540)

with open('line_with_spacecharge_and_particle.json', 'w') as fid:
    json.dump({
        'line': line.to_dict(),
        'particle': part.to_dict()},
        fid, cls=xo.JEncoder, indent=4)



import numpy as np

import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_custom_setter_array_element(test_context):
    line = xt.Line(elements=dict(
                qf1=xt.Multipole(knl=[0, 0.1]),
                qf2=xt.Multipole(knl=[0, 0.2]),
                qf3=xt.Multipole(knl=[0, 0.3]),
                dr=xt.Drift(length=1.)),
            element_names=['qf1', 'qf2', 'qf3', 'dr'])

    line.build_tracker(_context=test_context)

    elements_to_trim = [nn for nn in line.element_names if nn.startswith('qf')]

    qf_setter = xt.MultiSetter(line, elements_to_trim,
                                field='knl', index=1 # we want to change knl[1]
                                )

    ctx2np = test_context.nparray_from_context_array

    values = qf_setter.get_values()
    assert np.all(ctx2np(values) == np.array([0.1, 0.2, 0.3]))

    qf_setter.set_values(np.array([10., 100., 1000.]))
    assert np.all(ctx2np(qf_setter.get_values()) == np.array([10., 100., 1000.]))
    assert line['qf1'].knl[1] == 10.
    assert line['qf2'].knl[1] == 100.
    assert line['qf3'].knl[1] == 1000.


@for_all_test_contexts
def test_custom_setter_scalar(test_context):
    line = xt.Line(elements=dict(
                qf1=xt.Multipole(knl=[0, 1], length=0.1),
                qf2=xt.Multipole(knl=[0, 1], length=0.2),
                qf3=xt.Multipole(knl=[0, 1], length=0.3),
                dr=xt.Drift(length=1.)),
            element_names=['qf1', 'qf2', 'qf3', 'dr'])

    line.build_tracker(_context=test_context)

    elements_to_trim = [nn for nn in line.element_names if nn.startswith('qf')]

    qf_setter = xt.MultiSetter(line, elements_to_trim,
                                field='length',
                                )

    ctx2np = test_context.nparray_from_context_array

    values = qf_setter.get_values()
    assert np.all(ctx2np(values) == np.array([0.1, 0.2, 0.3]))

    qf_setter.set_values(np.array([10., 100., 1000.]))
    assert np.all(ctx2np(qf_setter.get_values()) == np.array([10., 100., 1000.]))
    assert line['qf1'].length == 10.
    assert line['qf2'].length == 100.
    assert line['qf3'].length == 1000.


@for_all_test_contexts
def test_custom_setter_fast_quadrupole(test_context):
    line = xt.Line(elements=dict(
                qf1=xt.SimpleThinQuadrupole(knl=[0, 1]),
                qf2=xt.SimpleThinQuadrupole(knl=[0, 1]),
                qf3=xt.SimpleThinQuadrupole(knl=[0, 1]),
                dr=xt.Drift(length=1.)),
            element_names=['qf1', 'qf2', 'qf3', 'dr'])

    line.build_tracker(_context=test_context)

    elements_to_trim = [nn for nn in line.element_names if nn.startswith('qf')]

    qf_setter = xt.MultiSetter(line, elements_to_trim,
                               field='knl', index=1)

    ctx2np = test_context.nparray_from_context_array

    values = qf_setter.get_values()
    assert np.all(ctx2np(values) == np.array([1, 1, 1]))

    qf_setter.set_values(np.array([1.1, 1.2, 1.3]))
    assert np.all(ctx2np(qf_setter.get_values()) == np.array([1.1, 1.2, 1.3]))
    assert line['qf1'].knl[1] == 1.1
    assert line['qf2'].knl[1] == 1.2
    assert line['qf3'].knl[1] == 1.3


@for_all_test_contexts
def test_custom_setter_fast_dipole(test_context):
    line = xt.Line(elements=dict(
                df1=xt.SimpleThinBend(knl=[1], length=0.1, hxl=0.1),
                df2=xt.SimpleThinBend(knl=[1], length=0.2, hxl=0.1),
                df3=xt.SimpleThinBend(knl=[1], length=0.3, hxl=0.1),
                dr=xt.Drift(length=1.)),
            element_names=['df1', 'df2', 'df3', 'dr'])

    line.build_tracker(_context=test_context)

    elements_to_trim = [nn for nn in line.element_names if nn.startswith('df')]

    qf_setter = xt.MultiSetter(line, elements_to_trim,
                               field='knl', index=0)

    ctx2np = test_context.nparray_from_context_array

    values = qf_setter.get_values()
    assert np.all(ctx2np(values) == np.array([1, 1, 1]))

    qf_setter.set_values(np.array([1.1, 1.2, 1.3]))
    assert np.all(ctx2np(qf_setter.get_values()) == np.array([1.1, 1.2, 1.3]))
    assert line['df1'].knl[0] == 1.1
    assert line['df2'].knl[0] == 1.2
    assert line['df3'].knl[0] == 1.3


@for_all_test_contexts
def test_custom_setter_empty(test_context):
    line = xt.Line(elements={
            'dr1': xt.Drift(length=1.),
            'dr2': xt.Drift(length=1.),
            'dr3': xt.Drift(length=1.),
        },
        element_names=['dr1', 'dr2', 'dr3', 'dr1'],
    )

    line.build_tracker(_context=test_context)
    qf_setter = xt.MultiSetter(line, [], field='knl', index=0)

    ctx2np = test_context.nparray_from_context_array

    values = qf_setter.get_values()
    assert not np.size(ctx2np(values))



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# We consider a case in which all RF cavities are off
tab = line.get_table()
tab_cav = tab.rows[tab.element_type == 'Cavity']
for nn in tab_cav.name:
    line[nn].voltage = 0

# For this configuration, `line.twiss()` gives an error because the
# longitudinal motion is not stable.
# In this case, the '4d' method of `line.twiss()` can be used to compute the
# twiss parameters.

tw = line.twiss(method='4d')

#!end-doc-part

import numpy as np

###########################################
# Match a particle distribution (4d mode) #
###########################################

# The '4d' method can also be used to match a particle distribution:

particles = line.build_particles(method='4d',
                    x_norm=[1,2,3,4], # sigmas
                    nemitt_x=2.5e-6, nemitt_y=2.5e-6)

#!end-doc-part

import matplotlib.pyplot as plt

plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(tw['s'], tw['betx'])
spbet.plot(tw['s'], tw['bety'])

spco.plot(tw['s'], tw['x'])
spco.plot(tw['s'], tw['y'])

spdisp.plot(tw['s'], tw['dx'])
spdisp.plot(tw['s'], tw['dy'])

spbet.set_ylabel(r'$\beta_{x,y}$ [m]')
spco.set_ylabel(r'(Closed orbit)$_{x,y}$ [m]')
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel('s [m]')

fig1.suptitle(
    r'$q_x$ = ' f'{tw["qx"]:.5f}' r' $q_y$ = ' f'{tw["qy"]:.5f}' '\n'
    r"$Q'_x$ = " f'{tw["dqx"]:.2f}' r" $Q'_y$ = " f'{tw["dqy"]:.2f}'
    r' $\gamma_{tr}$ = '  f'{1/np.sqrt(tw["momentum_compaction_factor"]):.2f}'
)

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)
plt.show()

#!end-doc-part


import numpy as np
import xobjects as xo
import xtrack as xt

class LinearRescale():

    def __init__(self, knob_name, v0, dv):
            self.knob_name = knob_name
            self.v0 = v0
            self.dv = dv

def _footprint_with_linear_rescale(linear_rescale_on_knobs, line,
                                   freeze_longitudinal=False,
                                   delta0=None, zeta0=None, 
                                   kwargs={}):

        if isinstance (linear_rescale_on_knobs, LinearRescale):
            linear_rescale_on_knobs = [linear_rescale_on_knobs]

        assert len(linear_rescale_on_knobs) == 1, (
            'Only one linear rescale is supported for now')

        knobs_0 = {}
        for rr in linear_rescale_on_knobs:
            nn = rr.knob_name
            v0 = rr.v0
            knobs_0[nn] = v0

        with xt._temp_knobs(line, knobs_0):
            fp = line.get_footprint(
                freeze_longitudinal=freeze_longitudinal,
                delta0=delta0, zeta0=zeta0, **kwargs)

        qx0 = fp.qx
        qy0 = fp.qy

        for rr in linear_rescale_on_knobs:
            nn = rr.knob_name
            v0 = rr.v0
            dv = rr.dv

            knobs_1 = knobs_0.copy()
            knobs_1[nn] = v0 + dv

            with xt._temp_knobs(line, knobs_1):
                fp1 = line.get_footprint(freeze_longitudinal=freeze_longitudinal,
                                        delta0=delta0, zeta0=zeta0, **kwargs)
            delta_qx = (fp1.qx - qx0) / dv * (line.vars[nn]._value - v0)
            delta_qy = (fp1.qy - qy0) / dv * (line.vars[nn]._value - v0)

            fp.qx += delta_qx
            fp.qy += delta_qy

        return fp

class Footprint():

    def __init__(self, nemitt_x=None, nemitt_y=None, n_turns=256, n_fft=2**18,
            mode='polar', r_range=None, theta_range=None, n_r=None, n_theta=None,
            x_norm_range=None, y_norm_range=None, n_x_norm=None, n_y_norm=None,
            keep_fft=False, keep_tracking_data=False,
            auto_to_numpy=True,fft_chunk_size=200
            ):

        assert nemitt_x is not None and nemitt_y is not None, (
            'nemitt_x and nemitt_y must be provided')
        self.mode = mode
        self.auto_to_numpy = auto_to_numpy
        self.n_turns = n_turns
        self.n_fft = n_fft
        self.fft_chunk_size = fft_chunk_size
        self.keep_fft = keep_fft
        self.keep_tracking_data = keep_tracking_data

        self.nemitt_x = nemitt_x
        self.nemitt_y = nemitt_y

        assert mode in ['polar', 'uniform_action_grid'], (
            'mode must be either polar or uniform_action_grid')

        if mode == 'polar':

            assert x_norm_range is None and y_norm_range is None, (
                'x_norm_range and y_norm_range must be None for mode polar')
            assert n_x_norm is None and n_y_norm is None, (
                'n_x_norm and n_y_norm must be None for mode polar')

            if r_range is None:
                r_range = (0.1, 6)
            if theta_range is None:
                theta_range = (0.05, np.pi/2-0.05)
            if n_r is None:
                n_r = 10
            if n_theta is None:
                n_theta = 10

            self.r_range = r_range
            self.theta_range = theta_range
            self.n_r = n_r
            self.n_theta = n_theta

            self.r_grid = np.linspace(*r_range, n_r)
            self.theta_grid = np.linspace(*theta_range, n_theta)
            self.R_2d, self.Theta_2d = np.meshgrid(self.r_grid, self.theta_grid)

            self.x_norm_2d = self.R_2d * np.cos(self.Theta_2d)
            self.y_norm_2d = self.R_2d * np.sin(self.Theta_2d)

        elif mode == 'uniform_action_grid':

            assert r_range is None and theta_range is None, (
                'r_range and theta_range must be None for mode uniform_action_grid')
            assert n_r is None and n_theta is None, (
                'n_r and n_theta must be None for mode uniform_action_grid')

            if x_norm_range is None:
                x_norm_range = (0.1, 6)
            if y_norm_range is None:
                y_norm_range = (0.1, 6)
            if n_x_norm is None:
                n_x_norm = 10
            if n_y_norm is None:
                n_y_norm = 10

            Jx_min = nemitt_x * x_norm_range[0]**2 / 2
            Jx_max = nemitt_x * x_norm_range[1]**2 / 2
            Jy_min = nemitt_y * y_norm_range[0]**2 / 2
            Jy_max = nemitt_y * y_norm_range[1]**2 / 2

            self.Jx_grid = np.linspace(Jx_min, Jx_max, n_x_norm)
            self.Jy_grid = np.linspace(Jy_min, Jy_max, n_y_norm)

            self.Jx_2d, self.Jy_2d = np.meshgrid(self.Jx_grid, self.Jy_grid)

            self.x_norm_2d = np.sqrt(2 * self.Jx_2d / nemitt_x)
            self.y_norm_2d = np.sqrt(2 * self.Jy_2d / nemitt_y)

    def _compute_footprint(self, line, freeze_longitudinal=False,
                           delta0=None, zeta0=None):

        if freeze_longitudinal is None:
            # In future we could detect if the line has frozen longitudinal plane
            freeze_longitudinal = False

        nplike_lib = line._context.nplike_lib

        particles = line.build_particles(
            x_norm=self.x_norm_2d.flatten(), y_norm=self.y_norm_2d.flatten(),
            nemitt_x=self.nemitt_x, nemitt_y=self.nemitt_y,
            zeta=zeta0, delta=delta0,
            freeze_longitudinal=freeze_longitudinal,
            method={True: '4d', False: '6d'}[freeze_longitudinal]
            )

        print('Tracking particles for footprint...')
        line.track(particles, num_turns=self.n_turns, turn_by_turn_monitor=True,
                   freeze_longitudinal=freeze_longitudinal)
        print('Done tracking.')

        ctx2np = line._context.nparray_from_context_array
        assert np.all(ctx2np(particles.state == 1)), (
            'Some particles were lost during tracking')
        mon = line.record_last_track
        mon.auto_to_numpy = False

        if isinstance(line._context, xo.ContextPyopencl):
            raise NotImplementedError(
                'Footprint calculation with Pyopencl not supported yet. '
                'Let us know if you need this feature.')
            # Could be implemented using xobject fft

        x_noCO = mon.x - nplike_lib.atleast_2d(mon.x.mean(axis=1)).T
        y_noCO = mon.y - nplike_lib.atleast_2d(mon.y.mean(axis=1)).T

        freq_axis = nplike_lib.fft.rfftfreq(self.n_fft)

        npart = nplike_lib.shape(x_noCO)[0]
        self.qx = nplike_lib.zeros(npart,dtype=float)
        self.qy = nplike_lib.zeros(npart,dtype=float)

        if self.keep_fft:
            self.fft_x = nplike_lib.zeros((npart,len(freq_axis)),dtype=complex)
            self.fft_y = nplike_lib.zeros((npart,len(freq_axis)),dtype=complex)

        # Compute in chunks
        iStart = 0
        while iStart < npart:
            iEnd = iStart + self.fft_chunk_size
            if iEnd > npart:
                iEnd = npart
            fft_x = nplike_lib.fft.rfft(x_noCO[iStart:iEnd,:], n=self.n_fft)
            fft_y = nplike_lib.fft.rfft(y_noCO[iStart:iEnd,:], n=self.n_fft)
            if self.keep_fft:
                self.fft_x[iStart:iEnd,:] = fft_x
                self.fft_y[iStart:iEnd,:] = fft_y
            qx = freq_axis[nplike_lib.argmax(nplike_lib.abs(fft_x), axis=1)]
            qy = freq_axis[nplike_lib.argmax(nplike_lib.abs(fft_y), axis=1)]
            self.qx[iStart:iEnd] = qx
            self.qy[iStart:iEnd] = qy
            iStart += self.fft_chunk_size

        self.qx = nplike_lib.reshape(self.qx, self.x_norm_2d.shape)
        self.qy = nplike_lib.reshape(self.qy, self.y_norm_2d.shape)

        if self.auto_to_numpy:
            ctx2np = line._context.nparray_from_context_array
            self.qx = ctx2np(self.qx)
            self.qy = ctx2np(self.qy)
            if self.keep_fft:
                self.fft_x = ctx2np(self.fft_x)
                self.fft_y = ctx2np(self.fft_y)

        if self.keep_tracking_data:
            self.tracking_data = mon

        print ('Done computing footprint.')

    def _compute_tune_shift(self,_context,J1_2d,J1_grid,J2_2d,J2_grid,q,coherent_tune,epsilon):
        nplike_lib = _context.nplike_lib
        ctx2np = _context.nparray_from_context_array
        np2ctx = _context.nparray_to_context_array

        integrand = -J1_2d*nplike_lib.exp(-J1_2d-J2_2d) / (coherent_tune - q + epsilon*1j)
        tune_shift = ctx2np(-1.0/nplike_lib.trapz(J2_grid,nplike_lib.trapz(J1_grid,integrand,1),0))
        return tune_shift

    def _compute_tune_shift_adaptive_epsilon(self,_context,J1_2d,J1_grid,J2_2d,J2_grid,q,coherent_tune,
                                             epsilon0,epsilon_factor,epsilon_rel_tol,max_iter,min_epsilon):
        tune_shift = self._compute_tune_shift(_context,J1_2d,J1_grid,J2_2d,J2_grid,q,coherent_tune,epsilon0)
        if epsilon_factor > 0.0:
            epsilon_ref = epsilon0
            epsilon = np.abs(np.imag(tune_shift)*epsilon_factor)
            if epsilon < min_epsilon:
                epsilon = min_epsilon
            count = 0
            while np.abs(1-epsilon/epsilon_ref) > epsilon_rel_tol and count < max_iter and epsilon >= min_epsilon:
                tune_shift = self._compute_tune_shift(_context,J1_2d,J1_grid,J2_2d,J2_grid,q,coherent_tune,epsilon)
                epsilon_ref = epsilon
                epsilon = np.abs(np.imag(tune_shift)*epsilon_factor)
                count += 1
        return tune_shift

    def get_stability_diagram(
        self,
        _context=None,
        n_points_stabiliy_diagram=100,
        epsilon0=1e-5,
        epsilon_factor=0.1,
        epsilon_rel_tol=0.1,
        max_iter=10,
        min_epsilon=1e-6,
        n_points_interpolate=1000,
    ):
        """
        Compute the stability diagram by evaluating the dispersion integral from [1]
        numerically for a set of complex tune shifts with vanishing imaginary part.
        By convention the imaginary part is positive.

        Parameters
        ----------
        _context:
        n_points_stabiliy_diagram: scalar(int)
            Number of times that the dispersion integral will be solved,
            each yielding a point on the output stability diagram
        epsilon0: scalar(float)
            vanishing imaginary part of the tune shift
        epsilon_factor: scalar(float)
            if larger than 0, an adaptive algorithm will be used to adjust
            epsilon between epsilon0 and epsilon_min using relative varitions
            in the order of the epsilon_factor
        epsilon_rel_tol: scalar(float)
            Stop the iterative algorithm if the relative change of
            epilson is smaller than epsilon_rel_tol
        max_iter: scalar(int)
            Stop the iterative algorithm if the the number of iterations
            reached max_iter
        min_epsilon: scalar(float)
            Stop the iterative algorithm if the epsilon is smaller than
            min_epsilon
        n_points_interpolate: scalar(int)
            Perform the numerical integration on a grid finer than the footprint
            using linear interpolation

        Returns
        -------
        tune_shifts_x: array_like(complex)
            Horizontal stability diagram
        tune_shifts_y: array_like(complex)
            Vertical stability diagram

        References
        ----------
        [1] https://cds.cern.ch/record/318826
        [2] https://doi.org/10.1103/PhysRevSTAB.17.111002
        """
        if _context == None:
            _context = xo.ContextCpu()
        nplike_lib = _context.nplike_lib
        splike_lib = _context.splike_lib
        ctx2np = _context.nparray_from_context_array
        np2ctx = _context.nparray_to_context_array

        Jx_2d = np2ctx(self.Jx_2d / self.nemitt_x)
        Jx_grid = np2ctx(self.Jx_grid / self.nemitt_x)
        Jy_2d = np2ctx(self.Jy_2d / self.nemitt_y)
        Jy_grid = np2ctx(self.Jy_grid / self.nemitt_y)
        qx = np2ctx(self.qx)
        qy = np2ctx(self.qy)

        if n_points_interpolate > len(Jx_grid) or n_points_interpolate > len(Jy_grid):
            interpolator_x = splike_lib.interpolate.RegularGridInterpolator(
                points=[Jy_grid, Jx_grid], values=qx, bounds_error=True, fill_value=None
            )
            interpolator_y = splike_lib.interpolate.RegularGridInterpolator(
                points=[Jy_grid, Jx_grid], values=qy, bounds_error=True, fill_value=None
            )
            Jx_grid = nplike_lib.linspace(Jx_grid[0], Jx_grid[-1], n_points_interpolate)
            Jy_grid = nplike_lib.linspace(Jy_grid[0], Jy_grid[-1], n_points_interpolate)
            Jx_2d, Jy_2d = nplike_lib.meshgrid(Jx_grid, Jy_grid)
            qx = interpolator_x((Jy_2d, Jx_2d))
            qy = interpolator_y((Jy_2d, Jx_2d))

        coherent_tunes_x = np.linspace(
            np.min(self.qx), np.max(self.qx), n_points_stabiliy_diagram
        )
        coherent_tunes_y = np.linspace(
            np.min(self.qy), np.max(self.qy), n_points_stabiliy_diagram
        )
        tune_shifts_x = np.zeros_like(coherent_tunes_x, dtype=complex)
        tune_shifts_y = np.zeros_like(coherent_tunes_y, dtype=complex)
        for i in range(n_points_stabiliy_diagram):
            tune_shifts_x[i] = self._compute_tune_shift_adaptive_epsilon(
                _context=_context,
                J1_2d=Jx_2d,
                J1_grid=Jx_grid,
                J2_2d=Jy_2d,
                J2_grid=Jy_grid,
                q=qx,
                coherent_tune=coherent_tunes_x[i],
                epsilon0=epsilon0,
                epsilon_factor=epsilon_factor,
                epsilon_rel_tol=epsilon_rel_tol,
                max_iter=max_iter,
                min_epsilon=min_epsilon,
            )
            tune_shifts_y[i] = self._compute_tune_shift_adaptive_epsilon(
                _context=_context,
                J1_2d=Jy_2d,
                J1_grid=Jy_grid,
                J2_2d=Jx_2d,
                J2_grid=Jx_grid,
                q=qy,
                coherent_tune=coherent_tunes_y[i],
                epsilon0=epsilon0,
                epsilon_factor=epsilon_factor,
                epsilon_rel_tol=epsilon_rel_tol,
                max_iter=max_iter,
                min_epsilon=min_epsilon,
            )
        return tune_shifts_x, tune_shifts_y

    def plot(self, ax=None, **kwargs):
        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()

        if 'color' not in kwargs:
            kwargs['color'] = 'k'

        labels = [None] * self.qx.shape[1]

        if 'label' in kwargs:
            label_str = kwargs['label']
            kwargs.pop('label')
            labels[0] = label_str

        ax.plot(self.qx, self.qy, label=labels, **kwargs)
        ax.plot(self.qx.T, self.qy.T, **kwargs)

        ax.set_xlabel(r'$q_x$')
        ax.set_ylabel(r'$q_y$')



import numpy as np

from .twiss import VARS_FOR_TWISS_INIT_GENERATION
from .general import _print, _LOC
import xtrack as xt
import xdeps as xd

XTRACK_DEFAULT_TOL = 1e-9
XTRACK_DEFAULT_SIGMA_REL = 0.01

XTRACK_DEFAULT_WEIGHTS = {
    # For quantities not specified here the default weight is 1
    'x': 10,
    'px': 100,
    'y': 10,
    'py': 100,
    'zeta': 10,
    'delta': 100,
    'pzeta': 100,
    'ptau': 100,
    'alfx': 10.,
    'alfy': 10.,
    'mux': 10.,
    'muy': 10.,
    'qx': 10.,
    'qy': 10.,
    'dx' : 10.,
    'dpx': 100.,
    'dy' : 10.,
    'dpy': 100.,
}

ALLOWED_TARGET_KWARGS= ['x', 'px', 'y', 'py', 'zeta', 'delta', 'pzata', 'ptau',
                        'betx', 'bety', 'alfx', 'alfy', 'gamx', 'gamy',
                        'mux', 'muy', 'dx', 'dpx', 'dy', 'dpy',
                        'qx', 'qy', 'dqx', 'dqy',
                        'ax_chrom', 'bx_chrom', 'ay_chrom', 'by_chrom',
                        'wx_chrom', 'wy_chrom',
                        'ddqx', 'ddqy', 'ddx', 'ddpx', 'ddy', 'ddpy',
                        'betx1', 'bety1', 'betx2', 'bety2',
                        'alfx1', 'alfy1', 'alfx2', 'alfy2',
                        'gamx1', 'gamy1', 'gamx2', 'gamy2',
                        'eq_gemitt_x', 'eq_gemitt_y', 'eq_gemitt_zeta',
                        'eq_nemitt_x', 'eq_nemitt_y', 'eq_nemitt_zeta']


# Alternative transitions functions
# def _transition_sigmoid_integral(x):
#     x_shift = x - 3
#     if x_shift > 10:
#         return x_shift
#     else:
#         return np.log(1 + np.exp(x_shift))

# def _transition_sin(x):
#     if x < 0:
#         return 0
#     if x < 1.:
#         return 2 /np.pi - 2 /np.pi * np.cos(np.pi * x / 2)
#     else:
#         return x + 2 / np.pi - 1

def _poly(x):
     return 3 * x**3 - 2 * x**4

def _transition_poly(x):
        x_cut = 1/16 + np.sqrt(33)/16
        if x < 0:
            return 0
        if x < x_cut:
            return _poly(x)
        else:
            return x - x_cut + _poly(x_cut)

class GreaterThan:

    _transition = staticmethod(_transition_poly)

    def __init__(self, lower, mode='step', sigma=None,
                 sigma_rel=XTRACK_DEFAULT_SIGMA_REL):
        assert mode in ['step', 'smooth']
        self.lower = lower
        self._value = 0.
        self.mode=mode
        if mode == 'smooth':
            assert sigma is not None or sigma_rel is not None
            if sigma is not None:
                assert sigma_rel is None
                self.sigma = sigma
            else:
                assert sigma_rel is not None
                lower_val = self.lower
                if xd.refs.is_ref(self.lower):
                    lower_val = self.lower._value
                self.sigma = np.abs(lower_val) * sigma_rel

    def auxtarget(self, res):
        '''Transformation applied to target value to obtain the corresponding
        cost function.
        '''
        if xd.refs.is_ref(self.lower):
            lower_val = self.lower._value
        else:
            lower_val = self.lower
        if self.mode == 'step':
            if res < lower_val:
                return res - lower_val
            else:
                return 0
        elif self.mode == 'smooth':
            return self.sigma * self._transition((lower_val - res) / self.sigma)
        elif self.mode == 'auxvar':
            raise NotImplementedError # experimental
            return res - lower_val - self.vary.container[self.vary.name]**2
        else:
            raise ValueError(f'Unknown mode {self.mode}')

    def __repr__(self):
        val = self.lower
        if xd.refs.is_ref(self.lower):
            val = self.lower._value
        return f'GreaterThan({val:.4g})'

    # Part of the `auxvar` experimental code
    # def _set_value(self, val, target):
    #     self.lower = val
    #     aux_vary_container = self.vary.container
    #     aux_vary_container[self.vary.name] = 0
    #     val = target.runeval()
    #     if val > 0:
    #         aux_vary_container[self.vary.name] = np.sqrt(val)
    # def gen_vary(self, container):
    #     self.vary = _gen_vary(container)
    #     return self.vary

class LessThan:

    _transition = staticmethod(_transition_poly)

    def __init__(self, upper, mode='step', sigma=None,
                 sigma_rel=XTRACK_DEFAULT_SIGMA_REL):
        assert mode in ['step', 'smooth']
        self.upper = upper
        self._value = 0.
        self.mode=mode
        if mode == 'smooth':
            assert sigma is not None or sigma_rel is not None
            if sigma is not None:
                assert sigma_rel is None
                self.sigma = sigma
            else:
                assert sigma_rel is not None
                upper_val = self.upper
                if xd.refs.is_ref(self.upper):
                    upper_val = self.upper._value
                self.sigma = np.abs(upper_val) * sigma_rel

    def auxtarget(self, res):
        if xd.refs.is_ref(self.upper):
            upper_val = self.upper._value
        else:
            upper_val = self.upper
        if self.mode == 'step':
            if res > upper_val:
                return upper_val - res
            else:
                return 0
        elif self.mode == 'smooth':
            return self.sigma * self._transition((res - upper_val) / self.sigma)
        elif self.mode == 'auxvar':
            raise NotImplementedError # experimental
            return upper_val - res - self.vary.container[self.vary.name]**2
        else:
            raise ValueError(f'Unknown mode {self.mode}')

    def __repr__(self):
        val = self.upper
        if xd.refs.is_ref(self.upper):
            val = self.upper._value
        return f'LessThan({val:.4g})'

# part of the `auxvar` experimental code
# def _gen_vary(container):
#     for ii in range(10000):
#         if f'auxvar_{ii}' not in container:
#             vv = f'auxvar_{ii}'
#             break
#     else:
#         raise RuntimeError('Too many auxvary variables')
#     container[vv] = 0
#     return xt.Vary(name=vv, container=container, step=1e-3)


class Target(xd.Target):

    def __init__(self, tar=None, value=None, at=None, tol=None, weight=None, scale=None,
                 line=None, action=None, tag=None, optimize_log=False,
                 **kwargs):

        """
        Target object for matching. Usage examples:

        .. code-block:: python

            Target('betx', 0.15, at='ip1', tol=1e-3)
            Target(betx=0.15, at='ip1', tol=1e-3)
            Target('betx', LessThan(0.15), at='ip1', tol=1e-3)
            Target('betx', GreaterThan(0.15), at='ip1', tol=1e-3)


        Parameters
        ----------
        tar : str or callable
            Name of the quantity to be matched or callable computing the
            quantity to be matched from the output of the action (by default the
            action is the Twiss action). Basic targets can also be specified
            using keyword arguments.
        value : float or xdeps.GreaterThan or xdeps.LessThan or xtrack.TwissTable
            Value to be matched. Inequality constraints can also be specified.
            If a TwissTable is specified, the value is obtained from the
            table using the specified tar and at.
        at : str, optional
            Element at which the quantity is evaluated. Needs to be specified
            if the quantity to be matched is not a scalar.
        tol : float, optional
            Tolerance below which the target is considered to be met.
        weight : float, optional
            Weight used for this target in the cost function.
        line : Line, optional
            Line in which the quantity is defined. Needs to be specified if the
            match involves multiple lines.
        action : Action, optional
            Action used to compute the quantity to be matched. By default the
            action is the Twiss action.
        tag : str, optional
            Tag associated to the target. Default is ''.
        optimize_log : bool, optional
            If True, the logarithm of the quantity is used in the cost function
            instead of the quantity itself. Default is False.
        """


        for kk in kwargs:
            assert kk in ALLOWED_TARGET_KWARGS, (
                f'Unknown keyword argument {kk}. '
                f'Allowed keywords are {ALLOWED_TARGET_KWARGS}')

        if len(kwargs) > 1:
            raise ValueError(f'{list(kwargs.keys())} cannot be specified '
                                'together in a single Target. Please use '
                                'multiple Targets or a TargetSet.')

        if len(kwargs) == 1:
            tar = list(kwargs.keys())[0]
            value = list(kwargs.values())[0]

        if at is not None:
            xdtar = (tar, at)
        else:
            xdtar = tar

        self._freeze_value = None

        if tag is None:
            tag_parts = []
            if line is not None:
                tag_parts.append(line)
            if at is not None:
                tag_parts.append(str(at))
            if isinstance(tar, str):
                tag_parts.append(tar)
            tag = '_'.join(tag_parts)

        xd.Target.__init__(self, tar=xdtar, value=value, tol=tol,
                            weight=weight, scale=scale, action=action, tag=tag,
                            optimize_log=optimize_log)
        self.line = line

    def __repr__(self):
        out = xd.Target.__repr__(self)
        if self.line is not None:
            lname = self.line
        elif hasattr(self.action, 'line') and hasattr(self.action.line , 'name'):
            lname = self.action.line.name
        else:
            lname = None
        if lname is not None:
            out = out.replace('Target(', f'Target(line={lname}, ')
        return out

    def eval(self, data):
        res = data[self.action]
        if self.line is not None:
            res = res[self.line]
        if callable(self.tar):
            out = self.tar(res)
        else:
            out = res[self.tar]

        if self._freeze_value is not None:
            return out

        return out

    def transform(self, val):
        if hasattr(self.value, 'auxtarget'):
            return self.value.auxtarget(val)
        else:
            return val

    @property
    def value(self):
        if self._freeze_value is not None:
            return self._freeze_value
        else:
            return self._user_value

    @value.setter
    def value(self, val):
        self._user_value = val

    def freeze(self):
        self._freeze_value = True # to bypass inequality logic
        self._freeze_value = self.runeval()

    def unfreeze(self):
        self._freeze_value = None

class TargetSet(xd.TargetList):

    def __init__(self, tars=None, value=None, at=None, tol=None, weight=None,
                 scale=None, line=None, action=None, tag=None, optimize_log=False,
                 **kwargs):

        """
        TargetSet object for matching, specifying a set of targets to be matched.

        Examples:

        .. code-block:: python
                TargetSet(['betx', 'bety'], 0.15, at='ip1', tol=1e-3)
                TargetSet(betx=0.15, bety=0.2, at='ip1', tol=1e-3)

        Parameters
        ----------
        tars : list, optional
            List of quantities to be matched. Basic targets can also be
            specified using keyword arguments.
        value : float or xdeps.GreaterThan or xdeps.LessThan
            Value to be matched. Inequality constraints can also be specified.
        at : str, optional
            Element at which the quantity is evaluated. Needs to be specified
            if the quantity to be matched is not a scalar.
        tol : float, optional
            Tolerance below which the target is considered to be met.
        weight : float, optional
            Weight used for this target in the cost function.
        line : Line, optional
            Line in which the quantity is defined. Needs to be specified if the
            match involves multiple lines.
        action : Action, optional
            Action used to compute the quantity to be matched. By default the
            action is the Twiss action.
        tag : str, optional
            Tag associated to the target. Default is ''.
        optimize_log : bool, optional
            If True, the logarithm of the quantity is used in the cost function
            instead of the quantity itself. Default is False.
        """

        if tars is not None and not isinstance(tars, (list, tuple)):
            tars = [tars]

        common_kwargs = locals().copy()
        common_kwargs.pop('self')
        common_kwargs.pop('kwargs')
        common_kwargs.pop('tars')
        common_kwargs.pop('value')

        vnames = []
        vvalues = []
        for kk in ALLOWED_TARGET_KWARGS:
            if kk in kwargs:
                vnames.append(kk)
                vvalues.append(kwargs[kk])
                kwargs.pop(kk)

        self.targets = []
        if tars is not None:
            self.targets += [Target(tt, value=value, **common_kwargs) for tt in tars]
        self.targets += [
            Target(tar=tar, value=val, **common_kwargs) for tar, val in zip(vnames, vvalues)]
        if len(self.targets) == 0:
            raise ValueError('No targets specified')

TargetList = TargetSet # for backward compatibility

class Vary(xd.Vary):

    def __init__(self, name, container=None, limits=None, step=None, weight=None,
                 max_step=None, active=True, tag=''):
        """
        Vary object for matching.

        Parameters
        ----------
        name : str
            Name of the variable to be varied.
        container : dict, optional
            Container in which the variable is defined. If not specified,
            line.vars is used.
        limits : tuple or None, optional
            Limits in which the variable is allowed to vary. Default is None.
        step : float, optional
            Step size used to compute the derivative of the cost function
            with respect to the variable.
        weight : float, optional
            Weight used for this vary in the cost function.
        max_step : float, optional
            Maximum allowed change in the variable per iteration.
        active : bool, optional
            Whether the variable is active in the optimization. Default is True.
        tag : str, optional
            Tag associated to the variable. Default is ''.

        """

        xd.Vary.__init__(self, name=name, container=container, limits=limits,
                         step=step, weight=weight, max_step=max_step, tag=tag,
                         active=active)

class VaryList(xd.VaryList):

    def __init__(self, vars, container=None, limits=None, step=None, weight=None,
                 max_step=None, active=True, tag=''):
        """
        VaryList object for matching specifying a list of variables to be varied.

        Parameters
        ----------
        vars : list
            List of variables to be varied.
        container : dict, optional
            Container in which the variables are defined. If not specified,
            line.vars is used.
        limits : tuple or None, optional
            Limits in which the variables are allowed to vary. Default is None.
        step : float, optional
            Step size used to compute the derivative of the cost function
            with respect to the variables.
        weight : float, optional
            Weight used for these variables in the cost function.
        max_step : float, optional
            Maximum allowed change in the variables per iteration.
        active : bool, optional
            Whether the variables are active in the optimization. Default is True.
        tag : str, optional
            Tag associated to the variables. Default is ''.
        """

        kwargs = dict(container=container, limits=limits, step=step,
                      weight=weight, max_step=max_step, active=active, tag=tag)
        self.vary_objects = [Vary(vv, **kwargs) for vv in vars]

class TargetInequality(Target):

    def __init__(self, tar, ineq_sign, rhs, at=None, tol=None, scale=None,
                 line=None, weight=None, tag=''):

        raise NotImplementedError('TargetInequality is not anymore supported. '
            'Please use Target with `GreaterThan` `LessThan` instead. '
            'For example, instead of '
            'TargetInequality("x", "<", 0.1, at="ip1") '
            'use '
            'Target("x", LessThan(0.1), at="ip1")')

class TargetRelPhaseAdvance(Target):

    def __init__(self, tar, value, end=None, start=None, tag='',  **kwargs):

        """
        Target object for matching the relative phase advance between two
        elements in a line computed as mu(end) - mu(start).

        Parameters
        ----------
        tar : str
            Phase advance to be matched. Can be either 'mux' or 'muy'.
        value : float or GreaterThan or LessThan or TwissTable
            Value to be matched. Inequality constraints can also be specified.
            If a TwissTable is specified, the target obtained from the table
            using the specified tar and at.
        end : str, optional
            Final element at which the phase advance is evaluated. Default is the
            last element of selected twiss range.
        start : str, optional
            Initali wlement at which the phase advance is evaluated. Default is the
            first element of the selected twiss range.
        tol : float, optional
            Tolerance below which the target is considered to be met.
        weight : float, optional
            Weight used for this target in the cost function.
        line : Line, optional
            Line in which the phase advance is defined. Needs to be specified if the
            match involves multiple lines.
        tag : str, optional
            Tag associated to the target. Default is ''.
        """

        Target.__init__(self, tar=self.compute, value=value, tag=tag, **kwargs)

        assert tar in ['mux', 'muy'], 'Only mux and muy are supported'
        self.var = tar
        if end is None:
            end = '__ele_stop__'
        if start is None:
            start = '__ele_start__'
        self.end = end
        self.start = start

    def __repr__(self):
        return f'TargetPhaseAdv({self.var}({self.end} - {self.start}), val={self.value}, tol={self.tol}, weight={self.weight})'

    def compute(self, tw):

        if self.end == '__ele_stop__':
            mu_1 = tw[self.var, -1]
        else:
            mu_1 = tw[self.var, self.end]

        if self.start == '__ele_start__':
            mu_0 = tw[self.var, 0]
        else:
            mu_0 = tw[self.var, self.start]

        return mu_1 - mu_0


class TargetRmatrixTerm(Target):

    def __init__(self, tar, value, start=None, end=None, tag='',  **kwargs):

        """
        Target object for matching terms of the R-matrix between two
        elements in a line.

        Parameters
        ----------
        tar : str
            Term to be matched. Can be "r11", "r12", "r21", "r22", etc
        value : float or GreaterThan or LessThan or TwissTable
            Value to be matched. Inequality constraints can also be specified.
            If a TwissTable is specified, the target obtained from the table
            using the specified tar and at.
        start : str
            First element of the range for which the R-matrix is computed.
        end : str
            End element of the range for which the R-matrix is computed.
        tol : float, optional
            Tolerance below which the target is considered to be met.
        weight : float, optional
            Weight used for this target in the cost function.
        line : Line, optional
            Line in which the R matrix is calculated. Needs to be specified if the
            match involves multiple lines.
        tag : str, optional
            Tag associated to the target. Default is ''.
        """

        assert isinstance(tar, str), 'Only strings are supported for tar'
        assert len(tar) == 3, (
            'Only terms of the R-matrix in the form "r11", "r12", "r21", "r22", etc'
            ' are supported')

        Target.__init__(self, tar=self.compute, value=value, tag=tag, **kwargs)

        self.term = tar
        if end is None:
            raise NotImplementedError('end cannot be None')
            end = '__ele_stop__'
        if start is None:
            raise NotImplementedError('start cannot be None')
            start = '__ele_start__'
        self.end = end
        self.start = start

    def __repr__(self):
        return f'{self.term}({self.start}, {self.end}, val={self.value}, tol={self.tol}, weight={self.weight})'

    def compute(self, tw):

        assert isinstance(self.term, str), 'Only strings are supported for tar'
        assert len(self.term) == 3, (
            'Only terms of the R-matrix in the form "r11", "r12", "r21", "r22", etc'
            ' are supported')

        if self.start is xt.START:
            self.start = tw.name[0]

        if self.end is xt.END:
            self.end = '_end_point'

        rmat = tw.get_R_matrix(self.start, self.end)

        ii = int(self.term[1]) - 1
        jj = int(self.term[2]) - 1

        assert ii >= 0 and ii < 6, 'Invalid R-matrix term'
        assert jj >= 0 and jj < 6, 'Invalid R-matrix term'

        return rmat[ii, jj]

class TargetRmatrix(TargetSet):

    def __init__(self, tars=None, value=None,start=None, end=None,
        r11=None, r12=None, r13=None, r14=None, r15=None, r16=None,
        r21=None, r22=None, r23=None, r24=None, r25=None, r26=None,
        r31=None, r32=None, r33=None, r34=None, r35=None, r36=None,
        r41=None, r42=None, r43=None, r44=None, r45=None, r46=None,
        r51=None, r52=None, r53=None, r54=None, r55=None, r56=None,
        r61=None, r62=None, r63=None, r64=None, r65=None, r66=None,
        **kwargs):

        if tars is not None:
            raise NotImplementedError
        if value is not None:
            raise NotImplementedError

        tag = kwargs.pop('tag', None)

        r_elems = {
            'r11': r11, 'r12': r12, 'r13': r13, 'r14': r14, 'r15': r15, 'r16': r16,
            'r21': r21, 'r22': r22, 'r23': r23, 'r24': r24, 'r25': r25, 'r26': r26,
            'r31': r31, 'r32': r32, 'r33': r33, 'r34': r34, 'r35': r35, 'r36': r36,
            'r41': r41, 'r42': r42, 'r43': r43, 'r44': r44, 'r45': r45, 'r46': r46,
            'r51': r51, 'r52': r52, 'r53': r53, 'r54': r54, 'r55': r55, 'r56': r56,
            'r61': r61, 'r62': r62, 'r63': r63, 'r64': r64, 'r65': r65, 'r66': r66,
        }

        tol = kwargs.pop('tol', None)

        self.targets = []
        for kk, vv in r_elems.items():
            thistol = tol
            if thistol is not None:
                if kk[1] in ['2', '4']:
                    thistol *= 1e-2
                if kk[2] in ['2', '4']:
                    thistol *= 1e+2
            if vv is not None:
                if tag is not None:
                    this_tag = tag
                else:
                    this_tag = kk
                self.targets.append(TargetRmatrixTerm(kk, vv, start=start, end=end,
                                                      tol=thistol, tag=this_tag,
                                                      **kwargs))


def match_line(line, vary, targets, solve=True, assert_within_tol=True,
                  compensate_radiation_energy_loss=False,
                  solver_options={}, allow_twiss_failure=True,
                  restore_if_fail=True, verbose=False,
                  n_steps_max=20, default_tol=None,
                  solver=None, check_limits=True,
                  name="",
                  **kwargs):

    opt = OptimizeLine(line, vary, targets,
                        assert_within_tol=assert_within_tol,
                        compensate_radiation_energy_loss=compensate_radiation_energy_loss,
                        solver_options=solver_options,
                        allow_twiss_failure=allow_twiss_failure,
                        restore_if_fail=restore_if_fail, verbose=verbose,
                        n_steps_max=n_steps_max, default_tol=default_tol,
                        solver=solver, check_limits=check_limits,
                        name=name,
                        **kwargs)

    if solve:
        opt.solve()

    return opt


class Action(xd.Action):

    _target_class = Target

    def __init__(self, callable, **kwargs):
        self.callable = callable
        self.kwargs = kwargs

    def run(self, allow_failure=False):
        return self.callable(**self.kwargs)

    def __call__(self, **kwargs):
        return self.run(**kwargs)

class ActionTwiss(xd.Action):

    def __init__(self, line, allow_twiss_failure=False,
                 compensate_radiation_energy_loss=False,
                 **kwargs):
        self.line = line
        self.kwargs = kwargs
        self.allow_twiss_failure = allow_twiss_failure
        self.compensate_radiation_energy_loss = compensate_radiation_energy_loss
        self._alredy_prepared = False

    def prepare(self, force=False):

        if self._alredy_prepared and not force:
            return

        line = self.line
        kwargs = self.kwargs

        ismultiline = isinstance(line, (xt.Multiline, xt.Environment, xt.MultilineLegacy))

        # Forbit specifying init through kwargs for Multiline
        if ismultiline:
            for kk in VARS_FOR_TWISS_INIT_GENERATION:
                if kk in kwargs:
                    raise ValueError(
                        f'`{kk}` cannot be specified for a Multiline match. '
                        f'Please specify provide a TwissInit object for each line instead.')

        # Handle init from table
        if ismultiline:

            line_names = kwargs.get('lines', line.line_names)
            none_list = [None] * len(line_names)
            twinit_list = kwargs.get('init', none_list)
            ele_start_list = kwargs.get('start', none_list)
            ele_stop_list = kwargs.get('end', none_list)

            assert isinstance(twinit_list, list)
            assert isinstance(ele_start_list, list)
            assert isinstance(ele_stop_list, list)

            for ii, twinit in enumerate(twinit_list):
                if isinstance(twinit, xt.MultiTwiss):
                    twinit_list[ii] = twinit[line_names[ii]]
        else:
            twinit_list = [kwargs.get('init', None)]
            ele_start_list = [kwargs.get('start', None)]
            ele_stop_list = [kwargs.get('end', None)]

            for ii, twinit in enumerate(twinit_list):
                if isinstance(twinit, xt.TwissInit):
                    twinit_list[ii] = twinit.copy()
                elif isinstance(twinit, str):
                    assert twinit == 'periodic' or twinit == 'periodic_symmetric'

        if ismultiline:
            kwargs['init'] = twinit_list
            kwargs['_keep_initial_particles'] = len(line_names) * [True]
        else:
            kwargs['init'] = twinit_list[0]
            kwargs['_keep_initial_particles'] = True

        tw0 = line.twiss(**kwargs)
        self._tw0 = tw0

        if ismultiline:
            kwargs['_initial_particles'] = len(line_names) * [None]
            for ii, llnn in enumerate(line_names):
                self.kwargs['init'][ii] = tw0[llnn].completed_init
                if not tw0[llnn].periodic:
                    kwargs['_initial_particles'][ii] = tw0[llnn]._data.get('_initial_particles', None)
        else:
            self.kwargs['init'] = tw0.completed_init
            for kk in VARS_FOR_TWISS_INIT_GENERATION:
                kwargs.pop(kk, None)
            if not(tw0.periodic):
                kwargs['_initial_particles'] = tw0._data.get(
                                        '_initial_particles', None)

        self.kwargs = kwargs

    def run(self, allow_failure=True):
        if self.compensate_radiation_energy_loss:
            if isinstance(self.line, (xt.Multiline, xt.Environment, xt.MultilineLegacy)):
                raise NotImplementedError(
                    'Radiation energy loss compensation is not yet supported'
                    ' for Multiline')
            self.line.compensate_radiation_energy_loss(verbose=False)
        if not self.allow_twiss_failure or not allow_failure:
            out = self.line.twiss(**self.kwargs)
        else:
            try:
                out = self.line.twiss(**self.kwargs)
            except Exception as ee:
                if allow_failure:
                    return 'failed'
                else:
                    raise ee
        out.line = self.line
        return out

class OptimizeLine(xd.Optimize):

    def __init__(self, line, vary, targets, assert_within_tol=True,
                    compensate_radiation_energy_loss=False,
                    solver_options={}, allow_twiss_failure=True,
                    restore_if_fail=True, verbose=False,
                    n_steps_max=20, default_tol=None,
                    solver=None, check_limits=True,
                    action_twiss=None,
                    name="",
                    **kwargs):

        if hasattr(targets, 'values'): # dict like
            targets = list(targets.values())

        if not isinstance(targets, (list, tuple)):
            targets = [targets]

        targets_flatten = []
        for tt in targets:
            if isinstance(tt, xd.TargetList):
                for tt1 in tt.targets:
                    targets_flatten.append(tt1.copy())
            else:
                targets_flatten.append(tt.copy())

        aux_vary = []

        for tt in targets_flatten:

            # Handle action
            if tt.action is None:
                if action_twiss is None:
                    action_twiss = ActionTwiss(
                        line, allow_twiss_failure=allow_twiss_failure,
                        compensate_radiation_energy_loss=compensate_radiation_energy_loss,
                        **kwargs)
                    action_twiss.prepare()
                tt.action = action_twiss

            # Handle at
            if isinstance(tt.tar, tuple):
                tt_name = tt.tar[0] # `at` is  present
                tt_at = tt.tar[1]
            else:
                tt_name = tt.tar
                tt_at = None
            if tt_at is not None and isinstance(tt_at, _LOC):
                assert isinstance(tt.action, ActionTwiss)
                tt.action.prepare() # does nothing if already prepared
                tw0 = tt.action._tw0[tt.line] if tt.line else tt.action._tw0
                this_line = tt.action.line[tt.line] if tt.line else tt.action.line
                if isinstance(tt_at, _LOC):
                    tt_at= tw0['name', {'START':0, 'END':-1}[tt_at.name]]
                    # If _end_point preceded by a marker, use the marker
                    if tt_at == '_end_point' and len(tw0.name) > 1:
                        nn_prev = tw0['name', -2]
                        nn_env_prev = tw0['name_env', -2]
                        if isinstance(this_line[nn_env_prev], xt.Marker):
                            tt_at= nn_prev
                tt.tar = (tt_name, tt_at)

            # Handle value
            if isinstance(tt.value, xt.MultiTwiss):
                tt.value=tt.value[tt.line][tt.tar]
            if isinstance(tt.value, xt.TwissTable):
                if isinstance(tt.tar, tuple) and tt.tar[1] == '_end_point':
                    # '_end_point' of the tar table might be different from the one of the action
                    raise ValueError('TwissTable target value cannot be used with at=_end_point')
                tt.value=tt.value[tt.tar]
            if isinstance(tt.value, np.ndarray):
                raise ValueError('Target value must be a scalar')

            # Handle weight
            if tt.weight is None:
                tt.weight = XTRACK_DEFAULT_WEIGHTS.get(tt_name, 1.)
            if tt.tol is None:
                if default_tol is None:
                    tt.tol = XTRACK_DEFAULT_TOL
                elif isinstance(default_tol, dict):
                    tt.tol = default_tol.get(tt_name,
                                        default_tol.get(None, XTRACK_DEFAULT_TOL))
                else:
                    tt.tol = default_tol

            # part of the `auxvar` experimental code
            # if isinstance(tt.value, (GreaterThan, LessThan)):
            #     if tt.value.mode == 'auxvar':
            #         aux_vary.append(tt.value.gen_vary(aux_vary_container))
            #         aux_vary_container[aux_vary[-1].name] = 0
            #         val = tt.runeval()
            #         if val > 0:
            #             aux_vary_container[aux_vary[-1].name] = np.sqrt(val)

        if not isinstance(vary, (list, tuple)):
            vary = [vary]

        vary = list(vary) + aux_vary

        vary_flatten = _flatten_vary(vary)
        _complete_vary_with_info_from_line(vary_flatten, line)

        xd.Optimize.__init__(self,
                        vary=vary_flatten, targets=targets_flatten, solver=solver,
                        verbose=verbose, assert_within_tol=assert_within_tol,
                        solver_options=solver_options,
                        n_steps_max=n_steps_max,
                        restore_if_fail=restore_if_fail,
                        check_limits=check_limits,
                        name=name)
        self.line = line
        self.action_twiss = action_twiss
        self.default_tol = default_tol

    def clone(self, add_targets=None, add_vary=None,
              remove_targets=None, remove_vary=None,
              name=None):

        if hasattr(add_targets, 'copy'):
            add_targets = add_targets.copy()

        if hasattr(add_vary, 'copy'):
            add_vary = add_vary.copy()

        if hasattr(add_targets, 'values'): # dict like
            add_targets = list(add_targets.values())

        if hasattr(add_vary, 'values'): # dict like
            add_vary = list(add_vary.values())

        if name is None:
            name = self.name

        assert remove_targets in [None, True, False]
        assert remove_vary in [None, True, False]

        targets = list(self.targets.copy())
        if remove_targets:
            targets = []
        if add_targets is not None:
            if not isinstance(add_targets, (list, tuple)):
                add_targets = [add_targets]
            targets.extend(add_targets)

        vary = list(self.vary.copy())
        if remove_vary:
            vary = []
        if add_vary is not None:
            if not isinstance(add_vary, (list, tuple)):
                add_vary = [add_vary]
            vary.extend(add_vary)

        out = self.__class__(
            line = self.line,
            vary=vary,
            targets=targets,
            default_tol=self.default_tol,
            restore_if_fail=self.restore_if_fail,
            verbose=self._err.verbose,
            assert_within_tol=self.assert_within_tol,
            n_steps_max=self.n_steps_max,
            check_limits=self.check_limits,
            action_twiss=self.action_twiss,
            name=name,
        )
        return out

    def plot(self, *args, **kwargs):
        return self.action_twiss.run().plot(*args, **kwargs)

def _flatten_vary(vary):
    vary_flatten = []
    for vv in vary:
        if isinstance(vv, xd.VaryList):
            for vv1 in vv.vary_objects:
                vary_flatten.append(vv1)
        else:
            vary_flatten.append(vv)
    return vary_flatten

def _complete_vary_with_info_from_line(vary, line):
    for vv in vary:
        if vv.container is None:
            vv.container = line.vars
            vv._complete_limits_and_step_from_defaults()

def closed_orbit_correction(line, line_co_ref, correction_config,
                            solver=None, verbose=False, restore_if_fail=True):

    for corr_name, corr in correction_config.items():
        _print('Correcting', corr_name)
        with xt.line._temp_knobs(line, corr['ref_with_knobs']):
            tw_ref = line_co_ref.twiss(method='4d', zeta0=0, delta0=0)
        vary = [xt.Vary(vv, step=1e-9) for vv in corr['vary']]
        targets = []
        for tt in corr['targets']:
            assert isinstance(tt, str), 'For now only strings are supported for targets'
            for kk in ['x', 'px', 'y', 'py']:
                targets.append(xt.Target(kk, at=tt, value=tw_ref[kk, tt], tol=1e-9))

        assert isinstance(corr['start'], str)

        line.match(
            solver=solver,
            verbose=verbose,
            restore_if_fail=restore_if_fail,
            vary=vary,
            targets=targets,
            init=xt.TwissInit(
                line=line,
                element_name=corr['start'],
                x=tw_ref['x', corr['start']],
                px=tw_ref['px', corr['start']],
                y=tw_ref['y', corr['start']],
                py=tw_ref['py', corr['start']],
                zeta=tw_ref['zeta', corr['start']],
                delta=tw_ref['delta', corr['start']],
            ),
            start=corr['start'], end=corr['end'])

def match_knob_line(line, knob_name, vary, targets, knob_value_start,
                    knob_value_end, run=True, **kwargs):

    knob_opt = KnobOptimizer(line, knob_name, vary, targets,
                    knob_value_start, knob_value_end,
                    **kwargs)
    if run:
        knob_opt.solve()
        knob_opt.generate_knob()
    return knob_opt

class KnobOptimizer:

    def __init__(self, line, knob_name, vary, targets,
                    knob_value_start, knob_value_end,
                    **kwargs):

        if not isinstance (vary, (list, tuple)):
            vary = [vary]

        vary_flatten = _flatten_vary(vary)
        _complete_vary_with_info_from_line(vary_flatten, line)

        vary_aux = []
        for vv in vary_flatten:
            aux_name = vv.name + '_from_' + knob_name
            if (aux_name in line.vars
                and (line.vars[aux_name] in
                         line.vars[vv.name]._expr._get_dependencies())):
                # reset existing term in expression
                line.vars[aux_name] = 0
            else:
                # create new term in expression
                line.vars[aux_name] = 0
                line.vars[vv.name] += line.vars[aux_name]

            vv_aux = vv.__dict__.copy()
            vv_aux['name'] = aux_name
            vary_aux.append(xt.Vary(**vv_aux))

        opt = line.match(vary=vary_aux, targets = targets, solve=False, **kwargs)

        object.__setattr__(self, 'opt', opt)
        self.line = line
        self.knob_name = knob_name
        self.knob_value_start = knob_value_start
        self.knob_value_end = knob_value_end

    def __getattr__(self, attr):
        return getattr(self.opt, attr)

    def __setattr__(self, attr, value):
        if hasattr(self.opt, attr):
            setattr(self.opt, attr, value)
        else:
            object.__setattr__(self, attr, value)

    def __dir__(self):
        return object.__dir__(self) + dir(self.opt)

    def generate_knob(self):
        self.line.vars[self.knob_name] = self.knob_value_end
        for vv in self.vary:
            var_value = self.line.vars[vv.name]._value
            self.line.vars[vv.name] = (
                var_value / (self.knob_value_end - self.knob_value_start)
                * (self.line.vars[self.knob_name]))
            if self.knob_value_start != 0:
                self.line.vars[vv.name] -= (
                    var_value / (self.knob_value_end - self.knob_value_start)
                    * self.knob_value_start)

        self.line.vars[self.knob_name] = self.knob_value_start

        _print('Generated knob: ', self.knob_name)

def opt_from_callable(function, x0, steps, tar, tols):

    '''Optimize a generic callable'''

    opt = xd.Optimize.from_callable(function, x0, tar, steps=steps, tols=tols,
                                    show_call_counter=False)
    return opt



'''
Adapted from P. Dijkstal, 2016, CERN
https://github.com/PyCOMPLETE/CellStudyInputPyECLOUD/

Based on A. Hofmann, The physics of synchrotron radiation
'''

from __future__ import division
import numpy as np
import scipy.special as special
import scipy.integrate as integrate
from scipy.constants import e, c, m_e, epsilon_0, h, hbar


def accept_arrays(func):
    def new_func(first_arg, *args, **kwargs):
        if hasattr(first_arg,'__len__'):
            return np.array([func(arg, *args, **kwargs) for arg in first_arg])
        else:
            return func(first_arg, *args, **kwargs)
    return new_func

def bessel_5_3(z):
    return special.kv(5./3., z)

@accept_arrays
def s_s(x):
    # Implements A. Hofmann, Eq. 5.17c
    factor = 9.*np.sqrt(3.) / (8*np.pi)
    upper_limit = np.inf
    integral, abs_error = integrate.quad(bessel_5_3, x, upper_limit)
    return factor * x * integral

def compute_power_single(gamma, rho):
    return e**2 / (4*np.pi*epsilon_0) * 2*c*gamma**4 / (3*rho**2)

def compute_critical_energy_eV(gamma, rho):
    return 3*c*h*gamma**3 / (4*np.pi*rho) / e

@accept_arrays
def integral(x_min, verbose=False):
    upper_limit = np.inf
    part_1, err_1 = integrate.quad(lambda x: bessel_5_3(x)*(x-x_min), x_min, upper_limit)
    if verbose:
        print('Absolute error for n_photons integral %.2e is %.2e' % (part_1, err_1))
    return part_1, err_1

@accept_arrays
def photon_spectrum(energy, gamma, rho):
    # Implements Hofmann, Eq. 5.48b
    ps = compute_power_single(gamma, rho)
    ec = compute_critical_energy_eV(gamma, rho)
    return ps / e /ec  / energy * s_s(energy/ec)




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.particle_ref = xt.Particles(
                    mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

line.vv['vrf400'] = 16
line.vv['kqsx3.l1'] = 3e-6

tw = line.twiss(method='6d')

cov = tw.get_beam_covariance(nemitt_x=2e-6, nemitt_y=1e-6, nemitt_zeta=5e-3)

Sig =  cov.Sigma[0]

S = xt.linear_normal_form.S

eival, eivec = np.linalg.eig(Sig @ S)

nex = eival[0].imag * tw.gamma0 * tw.beta0
ney = eival[1].imag * tw.gamma0 * tw.beta0
nez = eival[2].imag * tw.gamma0 * tw.beta0

print(f'{nex=} {ney=} {nez=}')

dummy_lam = np.diag([
    np.exp(-1j*np.pi/3), np.exp(+1j*np.pi/3),
    np.exp(-1j*np.pi/4), np.exp(+1j*np.pi/4),
    np.exp(-1j*np.pi/5), np.exp(+1j*np.pi/5),
])

dummy_R = eivec @ dummy_lam @ np.linalg.inv(eivec)

dummy_line = xt.Line(elements=[xt.Drift(length=1e-12)])
p_dummy = line.build_particles(x=0)

tw_from_sigmas = dummy_line.twiss(
                        particle_on_co=p_dummy,
                        R_matrix=dummy_R,
                        compute_chromatic_properties=False)

print('betx/bety')
print(f'betx (from line)    = {tw.betx[0]}')
print(f'betx (from sigmas)  = {tw_from_sigmas.betx[0]}')
print(f'bety (from line)    = {tw.bety[0]}')
print(f'bety (from sigmas)  = {tw_from_sigmas.bety[0]}')
print()
print('alfx/alfy')
print(f'alfx (from line)    = {tw.alfx[0]}')
print(f'alfx (from sigmas)  = {tw_from_sigmas.alfx[0]}')
print(f'alfy (from line)    = {tw.alfy[0]}')
print(f'alfy (from sigmas)  = {tw_from_sigmas.alfy[0]}')
print()
print('dx/dy')
print(f'dx (from line)      = {tw.dx[0]}')
print(f'dx (from sigmas)    = {tw_from_sigmas.dx[0]}')
print(f'dy (from line)      = {tw.dy[0]}')
print(f'dy (from sigmas)    = {tw_from_sigmas.dy[0]}')
print()
print('coupled betas')
print(f'betx2 (from line)   = {tw.betx2[0]}')
print(f'betx2 (from sigmas) = {tw_from_sigmas.betx2[0]}')
print(f'bety2 (from line)   = {tw.bety2[0]}')
print(f'bety2 (from sigmas) = {tw_from_sigmas.bety2[0]}')






import xtrack as xt
env = xt.get_environment()

env.vars.update(default_to_zero=True,

    # Quadrupole strengths
    qf1k1 =  3.15396e-01,
    qd1k1 = -5.24626e-01,
    qf2k1 =  5.22717e-01,

    # Chromaticity sextupole strengths
    k2xcf = -4.33238e-01,
    k2xcd = -5.52276e-01,

    # Extraction sextupole strengths
    k2xrr_a = 'k2xrr',
    k2xrr =  8.65,
)



import numpy as np
import xtrack as xt
import xobjects as xo

xo.context_default.kernels.clear()

model = 'full'

# only edge entry
bend_only_e1 = xt.Bend(
    length=0, k0=0.1,
    edge_entry_angle=0.05,
    edge_entry_hgap=0.06,
    edge_entry_fint=0.08,
    edge_exit_active=False)

edge_e1 = xt.DipoleEdge(
    k=0.1, side='entry', e1=0.05, hgap=0.06, fint=0.08)

line = xt.Line(elements=[bend_only_e1])

line['e0'].length = 1 # to force the slicing
line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Teapot(1))])
line['e0'].length = 0
line.build_tracker()

assert 'e0..entry_map' in line.element_names
assert 'e0..exit_map' in line.element_names
assert isinstance(line['e0..entry_map'], xt.ThinSliceBendEntry)
assert isinstance(line['e0..exit_map'], xt.ThinSliceBendExit)

edge_e1.model = model
bend_only_e1.edge_entry_model = model

p1 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03)
p2 = p1.copy()
p3 = p1.copy()

bend_only_e1.track(p1)
edge_e1.track(p2)
line.track(p3)

assert_allclose = np.testing.assert_allclose
assert_allclose(p1.x, p2.x, rtol=0, atol=1e-14)
assert_allclose(p1.px, p2.px, rtol=0, atol=1e-14)
assert_allclose(p1.y, p2.y, rtol=0, atol=1e-14)
assert_allclose(p1.py, p2.py, rtol=0, atol=1e-14)
assert_allclose(p1.zeta, p2.zeta, rtol=0, atol=1e-14)
assert_allclose(p1.delta, p2.delta, rtol=0, atol=1e-14)

assert_allclose(p1.x, p3.x, rtol=0, atol=1e-14)
assert_allclose(p1.px, p3.px, rtol=0, atol=1e-14)
assert_allclose(p1.y, p3.y, rtol=0, atol=1e-14)
assert_allclose(p1.py, p3.py, rtol=0, atol=1e-14)
assert_allclose(p1.zeta, p3.zeta, rtol=0, atol=1e-14)
assert_allclose(p1.delta, p3.delta, rtol=0, atol=1e-14)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from matplotlib import pyplot as plt

import xobjects as xo
import xtrack as xt

q_x_set = 0.18
q_y_set = 0.22
Q_s_set = 0.025

beta_x = 1.0
beta_y = 10.0
alpha_x = 0.0
alpha_y = 0.0
beta_s = 12.244
energy = 45.6

gamma_x = (1.0 + alpha_x**2) / beta_x
gamma_y = (1.0 + alpha_y**2) / beta_y

# radiation properties
damping_rate_h = 5e-4  # horizontal emittance damping rate
damping_rate_v = 1e-3  # vertical emittance damping rate
damping_rate_s = 2e-3  # longitudinal emittance damping rate
equ_emit_x = 0.3e-9
equ_emit_y = 1e-12
equ_length = 3.5e-3
equ_delta = 3.8e-4
beta_s = equ_length / equ_delta
equ_emit_s = equ_length * equ_delta

# Setting up diagonal terms of the damping and diffusion matrix based
# on radiation properties
damping_rate_x = damping_rate_h/2
damping_rate_px = damping_rate_h/2
damping_rate_y = damping_rate_v/2
damping_rate_py = damping_rate_v/2
damping_rate_pzeta = damping_rate_s

gauss_noise_ampl_px = np.sqrt(equ_emit_x*damping_rate_h/beta_x)
gauss_noise_ampl_x = beta_x*gauss_noise_ampl_px
gauss_noise_ampl_py = np.sqrt(equ_emit_y*damping_rate_v/beta_y)
gauss_noise_ampl_y = beta_y*gauss_noise_ampl_py
gauss_noise_ampl_delta = np.sqrt(2*equ_emit_s*damping_rate_s/beta_s)

# Build the map
ring_map = xt.LineSegmentMap(
    qx=q_x_set,
    qy=q_y_set,
    qs=Q_s_set,
    bets=beta_s,
    betx=beta_x,
    bety=beta_y,
    alfx=alpha_x,
    alfy=alpha_y,
    damping_rate_x=damping_rate_x,
    damping_rate_px=damping_rate_px,
    damping_rate_y=damping_rate_y,
    damping_rate_py=damping_rate_py,
    damping_rate_zeta=0.0,
    damping_rate_pzeta=damping_rate_pzeta,
    gauss_noise_ampl_x=gauss_noise_ampl_x,
    gauss_noise_ampl_px=gauss_noise_ampl_px,
    gauss_noise_ampl_y=gauss_noise_ampl_y,
    gauss_noise_ampl_py=gauss_noise_ampl_py,
    gauss_noise_ampl_zeta=0.0,
    gauss_noise_ampl_pzeta=gauss_noise_ampl_delta,
)

# Build particles (all at the same position)
num_particles = 1000
part = xt.Particles(
    x=num_particles * [2 * np.sqrt(equ_emit_x * beta_x)],
    y=num_particles * [2 * np.sqrt(equ_emit_y * beta_y)],
    pzeta=num_particles * [0.15 * np.sqrt(equ_delta / beta_s)],
    p0c=energy * 1e9,
)
# Initialize random number generator
part._init_random_number_generator()

# Build line and tracker
line = xt.Line(elements=[ring_map])
line.build_tracker()

# Track!
num_turns = 40_000
line.track(part, num_turns=num_turns,
           turn_by_turn_monitor=True, with_progress=True)

# Compute emittances from tracking data
emit_x = 0.5 * (
    gamma_x * line.record_last_track.x ** 2
    + 2 * alpha_x *
    line.record_last_track.x * line.record_last_track.px[0, :]
    + beta_x * line.record_last_track.px ** 2
).mean(axis=0)
emit_y = 0.5 * (
    gamma_y * line.record_last_track.y ** 2
    + 2 * alpha_y *
    line.record_last_track.y * line.record_last_track.py[0, :]
    + beta_y * line.record_last_track.py ** 2
).mean(axis=0)
emit_s = 0.5 * (line.record_last_track.zeta ** 2 /
                beta_s + beta_s * line.record_last_track.delta ** 2
                ).mean(axis=0)

# Get equilibrium emittances by averaging last turns
averga_start = 30_000
emit_x_0 = np.average(emit_x[averga_start:])
emit_y_0 = np.average(emit_y[averga_start:])
emit_s_0 = np.average(emit_s[averga_start:])

# Expected emittance evolution
turns = np.arange(num_turns)
eps_x_expected = (emit_x[0] - emit_x_0) * \
    np.exp(-damping_rate_h * turns) + emit_x_0
eps_y_expected = (emit_y[0] - emit_y_0) * \
    np.exp(-damping_rate_v * turns) + emit_y_0
eps_s_expected = (emit_s[0] - emit_s_0) * \
    np.exp(-damping_rate_s * turns) + emit_s_0

# Plot
plt.close("all")

plt.figure(1)
sp1 = plt.subplot(311)
plt.plot(turns, emit_x)
plt.plot(turns, eps_x_expected, "--r", label="expected")
plt.axhline(y=emit_x_0, linestyle='--', color='g', label="equilibrium")
plt.ylabel(r"$\epsilon_x$ [m]")
plt.ylim(bottom=0)
plt.legend()

plt.subplot(312, sharex=sp1)
plt.plot(turns, emit_y)
plt.plot(turns, eps_y_expected, "--r")
plt.plot([turns[0], turns[-1]], [emit_y_0, emit_y_0], "--g")
plt.ylim(bottom=0)
plt.ylabel(r"$\epsilon_y$ [m]")

plt.subplot(313, sharex=sp1)
plt.plot(turns, emit_s)
plt.plot(turns, eps_s_expected, "--r")
plt.plot([turns[0], turns[-1]], [emit_s_0, emit_s_0], "--g")
plt.ylabel(r"$\epsilon_s$ [m]")
plt.ylim(bottom=0)
plt.xlabel("turn")

plt.show()



import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe
from scipy.constants import epsilon_0 as eps0

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts
from xtrack._temp.boris_and_solenoid_map.solenoid_field import SolenoidField


@for_all_test_contexts
def test_solenoid_bz_map_vs_boris(test_context):
    ctx = xo.ContextCpu()

    boris_knl_description = xo.Kernel(
        c_name='boris_step',
        args=[
            xo.Arg(xo.Int64,   name='N_sub_steps'),
            xo.Arg(xo.Float64, name='Dtt'),
            xo.Arg(xo.Float64, name='B_field', pointer=True),
            xo.Arg(xo.Float64, name='B_skew', pointer=True),
            xo.Arg(xo.Float64, name='xn1', pointer=True),
            xo.Arg(xo.Float64, name='yn1', pointer=True),
            xo.Arg(xo.Float64, name='zn1', pointer=True),
            xo.Arg(xo.Float64, name='vxn1', pointer=True),
            xo.Arg(xo.Float64, name='vyn1', pointer=True),
            xo.Arg(xo.Float64, name='vzn1', pointer=True),
            xo.Arg(xo.Float64, name='Ex_n', pointer=True),
            xo.Arg(xo.Float64, name='Ey_n', pointer=True),
            xo.Arg(xo.Float64, name='Bx_n_custom', pointer=True),
            xo.Arg(xo.Float64, name='By_n_custom', pointer=True),
            xo.Arg(xo.Float64, name='Bz_n_custom', pointer=True),
            xo.Arg(xo.Int64,   name='custom_B'),
            xo.Arg(xo.Int64,   name='N_mp'),
            xo.Arg(xo.Int64,   name='N_multipoles'),
            xo.Arg(xo.Float64, name='charge'),
            xo.Arg(xo.Float64, name='mass', pointer=True),
        ],
    )

    ctx.add_kernels(
        kernels={'boris': boris_knl_description},
        sources=[xt._pkg_root / '_temp/boris_and_solenoid_map/boris.h'],
    )

    delta=np.array([0, 4])
    p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
                    energy0=45.6e9,
                    x=[-1e-3, -1e-3], px=-1e-3*(1+delta), y=1e-3,
                    delta=delta)

    p = p0.copy()

    sf = SolenoidField(L=4, a=0.3, B0=1.5, z0=20)

    dt = 1e-10
    n_steps = 1500

    x_log = []
    y_log = []
    z_log = []
    px_log = []
    py_log = []
    pp_log = []
    pow_log = []

    for ii in range(n_steps):

        x = p.x.copy()
        y = p.y.copy()
        z = p.s.copy()

        Bx, By, Bz = sf.get_field(x, y, z)

        gamma = p.energy / p.mass0
        mass0_kg = p.mass0 * qe / clight**2
        charge0_coulomb = p.q0 * qe

        p0c_J = p.p0c * qe

        Pxc_J = p.px * p0c_J
        Pyc_J = p.py * p0c_J
        Pzc_J = np.sqrt((p0c_J*(1 + p.delta))**2 - Pxc_J**2 - Pyc_J**2)

        vx = Pxc_J / clight / (gamma * mass0_kg) # m/s
        vy = Pyc_J / clight / (gamma * mass0_kg) # m/s
        vz = Pzc_J / clight / (gamma * mass0_kg) # m/s

        vx_before = vx.copy()
        vy_before = vy.copy()

        ctx.kernels.boris(
                N_sub_steps=1,
                Dtt=dt,
                B_field=np.array([0.]),
                B_skew=np.array([0.]),
                xn1=x,
                yn1=y,
                zn1=z,
                vxn1=vx,
                vyn1=vy,
                vzn1=vz,
                Ex_n=0 * x,
                Ey_n=0 * x,
                Bx_n_custom=Bx,
                By_n_custom=By,
                Bz_n_custom=Bz,
                custom_B=1,
                N_mp=len(x),
                N_multipoles=0,
                charge=charge0_coulomb,
                mass=mass0_kg * gamma,
        )

        p.x = x
        p.y = y
        p.s = z
        p.px = mass0_kg * gamma * vx * clight / p0c_J
        p.py = mass0_kg * gamma * vy * clight / p0c_J
        pz = mass0_kg * gamma * vz * clight / p0c_J
        pp = np.sqrt(p.px**2 + p.py**2 + pz**2)

        beta_x_before = vx_before / clight
        beta_y_before = vy_before / clight

        beta_x_after = vx / clight
        beta_y_after = vy / clight

        beta_x_dot = (beta_x_after - beta_x_before) / dt
        beta_y_dot = (beta_y_after - beta_y_before) / dt

        bet_dot_square = beta_x_dot**2 + beta_y_dot**2

        # From Hofmann, "The physics of synchrontron radiation" Eq 3.7 and below
        pow = 2 * qe**2 * bet_dot_square * gamma**4 / (12 * np.pi * eps0 * clight)

        x_log.append(p.x.copy())
        y_log.append(p.y.copy())
        z_log.append(p.s.copy())
        px_log.append(p.px.copy())
        py_log.append(p.py.copy())
        pp_log.append(pp)
        pow_log.append(pow)

    x_log = np.array(x_log)
    y_log = np.array(y_log)
    z_log = np.array(z_log)
    px_log = np.array(px_log)
    py_log = np.array(py_log)
    pp_log = np.array(pp_log)

    pow_log = np.array(pow_log)

    dE_ds_boris_J = 0 * pow_log

    dE_ds_boris_J[:-1]= pow_log[:-1] * dt / np.diff(z_log, axis=0)
    dE_ds_boris_eV = dE_ds_boris_J / qe

    z_axis = np.linspace(0, 30, 1001)
    Bz_axis = sf.get_field(0 * z_axis, 0 * z_axis, z_axis)[2]

    P0_J = p.p0c[0] * qe / clight
    brho = P0_J / qe / p.q0

    #ks = 0.5 * (Bz_axis[:-1] + Bz_axis[1:]) / brho
    ks = Bz_axis[:-1] / brho

    line = xt.Line(elements=[xt.Solenoid(length=z_axis[1]-z_axis[0], ks=ks[ii])
                                for ii in range(len(z_axis)-1)])
    line.build_tracker(_context=test_context)
    line.configure_radiation(model='mean')

    p_xt = p0.copy(_context=test_context)
    line.track(p_xt, turn_by_turn_monitor='ONE_TURN_EBE')
    mon = line.record_last_track

    p_xt = p0.copy(_context=test_context)
    line.configure_radiation(model=None)
    line.track(p_xt, turn_by_turn_monitor='ONE_TURN_EBE')
    mon_no_rad = line.record_last_track

    Bz_mid = 0.5 * (Bz_axis[:-1] + Bz_axis[1:])
    Bz_mon = 0 * Bz_axis
    Bz_mon[1:] = Bz_mid

    # Wolsky Eq. 3.114
    Ax = -0.5 * Bz_mon * mon.y
    Ay =  0.5 * Bz_mon * mon.x

    # Wolsky Eq. 2.74
    ax_ref = Ax * p0.q0 * qe / P0_J
    ay_ref = Ay * p0.q0 * qe / P0_J

    px_mech = mon.px - ax_ref
    py_mech = mon.py - ay_ref
    pz_mech = np.sqrt((1 + mon.delta)**2 - px_mech**2 - py_mech**2)

    dx_ds = np.diff(mon.x, axis=1) / np.diff(mon.s, axis=1)
    dy_ds = np.diff(mon.y, axis=1) / np.diff(mon.s, axis=1)

    ctx2np = p_xt._context.nparray_from_context_array
    dE_ds = -np.diff(mon.ptau, axis=1)/np.diff(mon.s, axis=1) * ctx2np(p_xt.energy0)[0]

    emitted_dpx = -(np.diff(mon.px, axis=1) - np.diff(mon_no_rad.px, axis=1))
    emitted_dpy = -(np.diff(mon.py, axis=1) - np.diff(mon_no_rad.py, axis=1))
    emitted_dp = -(np.diff(mon.delta, axis=1) - np.diff(mon_no_rad.delta, axis=1))

    z_check = sf.z0 + sf.L * np.linspace(-2, 2, 1001)

    for i_part in range(z_log.shape[1]):

        this_s_boris = 0.5 * (z_log[:-1, i_part] + z_log[1:, i_part])
        dx_ds_boris = np.diff(x_log[:, i_part]) / np.diff(z_log[:, i_part])
        dy_ds_boris = np.diff(y_log[:, i_part]) / np.diff(z_log[:, i_part])

        s_xsuite = 0.5 * (mon.s[i_part, :-1] + mon.s[i_part, 1:])
        dx_ds_xsuite = np.diff(mon.x[i_part, :]) / np.diff(mon.s[i_part, :])
        dy_ds_xsuite = np.diff(mon.y[i_part, :]) / np.diff(mon.s[i_part, :])
        dE_ds_xsuite = dE_ds[i_part, :]

        dx_ds_xsuite_check = np.interp(z_check, s_xsuite, dx_ds_xsuite)
        dy_ds_xsuite_check = np.interp(z_check, s_xsuite, dy_ds_xsuite)
        dE_ds_xsuite_check = np.interp(z_check, s_xsuite, dE_ds_xsuite)

        dx_ds_boris_check = np.interp(z_check, this_s_boris, dx_ds_boris)
        dy_ds_boris_check = np.interp(z_check, this_s_boris, dy_ds_boris)
        dE_ds_boris_check = np.interp(z_check, z_log[:, i_part], dE_ds_boris_eV[:, i_part])

        this_emitted_dpx = emitted_dpx[i_part, :]
        this_emitted_dpy = emitted_dpy[i_part, :]
        this_dE_ds = dE_ds[i_part, :]
        this_dx_ds = dx_ds[i_part, :]
        this_dy_ds = dy_ds[i_part, :]

        xo.assert_allclose(dx_ds_xsuite_check, dx_ds_boris_check, rtol=0,
                atol=2.8e-2 * (np.max(dx_ds_boris_check) - np.min(dx_ds_boris_check)))
        xo.assert_allclose(dy_ds_xsuite_check, dy_ds_boris_check, rtol=0,
                atol=2.8e-2 * (np.max(dy_ds_boris_check) - np.min(dy_ds_boris_check)))
        xo.assert_allclose(dE_ds_xsuite_check, dE_ds_boris_check, rtol=0,
                atol=1e-2 * (np.max(dE_ds_boris_check) - np.min(dE_ds_boris_check)))

        xo.assert_allclose(ax_ref[i_part, :], mon.ax[i_part, :],
                        rtol=0, atol=np.max(np.abs(ax_ref)*3e-2))
        xo.assert_allclose(ay_ref[i_part, :], mon.ay[i_part, :],
                        rtol=0, atol=np.max(np.abs(ay_ref)*3e-2))

        xo.assert_allclose(this_emitted_dpx,
                this_dE_ds * this_dx_ds * np.diff(mon.s[i_part, :])/p.p0c[0],
                rtol=0, atol=1e-4 * (np.max(this_emitted_dpx) - np.min(this_emitted_dpx)))
        xo.assert_allclose(this_emitted_dpy,
                this_dE_ds * this_dy_ds * np.diff(mon.s[i_part, :])/p.p0c[0],
                rtol=0, atol=2e-3 * (np.max(this_emitted_dpy) - np.min(this_emitted_dpy)))
